##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Objects.py,v 1.5.2.10 2009-07-24 13:35:53 ebke Exp $
##########################################################################
# NOTE: Make sure that _data and __dict__ of any GangaObject are only referenced
# here - this is necessary for write locking and lazy loading!

from functools import partial
import inspect
import abc
import threading
import _thread
from contextlib import contextmanager
import functools

from GangaCore.Utility.logging import getLogger

from copy import deepcopy, copy
from inspect import isclass

from GangaCore.GPIDev.Schema import Schema, Item, ComponentItem, SharedItem

from GangaCore.Core.exceptions import GangaValueError, GangaException

from GangaCore.Utility.Plugin import allPlugins

from GangaCore.Utility.Profiling import cpu_profiler, mem_profiler, call_counter

def _getName(obj):
    """ Return the name of an object based on what we prioritise 
    Name is defined as the first in the list of:
    obj._name, obj.__name__, obj.__class__.__name__ or str(obj)
    Args:
        obj (unknown):
    """
    try:
        return obj._name
    except AttributeError:
        try:
            return obj.__name__
        except AttributeError:
            try:
                return obj.__class__.__name__
            except AttributeError:
                return str(obj)

logger = getLogger()

do_not_copy = ['_index_cache_dict', '_parent', '_registry', '_data_dict', '_lock', '_proxyObject']

def synchronised(f):
    """
    This decorator must be attached to a method on a ``Node`` subclass
    It uses the object's lock to make sure that the object is held for the duration of the decorated function
    Args:
        f (function): This is the function which we want to wrap
    """
    @functools.wraps(f)
    def sync_decorated(self, *args, **kwargs):
        with self.const_lock:
            return f(self, *args, **kwargs)

    return sync_decorated


class Node(object, metaclass=abc.ABCMeta):
    """
    The Node class is the code of the Ganga heirachy. It allows objects to keep
    track of their parent, whether they're dirty and take part in the visitor
    pattern.

    It also provides access to tree-aware read/write locks to provide
    thread-safe usage.
    """
    __slots__ = ('_parent', '_lock', '_dirty')

    def __init__(self, parent=None):
        super(Node, self).__init__()
        self._parent = parent
        self._lock = threading.RLock()  # Don't write from out of thread when modifying an object
        self._dirty = False  # dirty flag is true if the object has been modified locally and its contents is out-of-sync with its repository

    def __deepcopy__(self, memo=None):
        cls = self.__class__
        obj = cls()
        this_dict = copy(self.__dict__)
        for elem, val in this_dict.items():
            if elem not in do_not_copy:
                this_dict[elem] = deepcopy(val, memo)

        obj.__dict__ = this_dict
        if self._getParent() is not None:
            obj._setParent(self._getParent())
        return obj

    def _getParent(self):
        """
        This gets the parent object defined for this Node
        """
        # type: () -> Node
        return self._parent

    @synchronised  # This will lock the _current_ (soon to be _old_) root object
    def _setParent(self, parent):
        """
        This sets the parent of this Node to be a new object
        Args:
            parent (GangaObject): This is the GangaObject to be taken as the objects new parent
        """
        # type: (Node) -> None
        if parent is None:
            setattr(self, '_parent', parent)
        else:
            with parent.const_lock: # This will lock the _new_ root object
                setattr(self, '_parent', parent)
            # Finally the new and then old root objects will be unlocked

    @property
    @contextmanager
    def const_lock(self):
        """
        This is a context manager which acquires the const write lock on the
        object's root object.

        This lock acquires exclusive access over an object tree to prevent it
        from changing. Reading schema attributes on the object is still allowed
        but changing them is not. Only one thread can hold this lock at once.
        """
        root = self._getRoot()
        with root._lock:
            yield

    def _getRoot(self, cond=None):
        # type: () -> Node
        """
        get the root of the object tree
        if parent does not exist then the root is the 'self' object
        cond is an optional function which may cut the search path: when it
        returns True, then the parent is returned as root
        """

        self_parent = self._getParent()
        if self_parent is None:
            return self
        obj = self_parent

        if cond is None:
            while True:
                parent = obj._getParent()
                if parent is not None:
                    obj = parent
                else:
                    break
                return obj
        else:
                                                    
            if cond(obj):
                return obj
                                                                                  
            escape = False
            while True:
                parent = obj._getObject()
                if parent is None or cond(obj):
                    break
                obj = parent
            return obj

        if cond is None:
            while obj._getParent() is not None:
                obj = obj._getParent()
            return obj
        else:

            if cond(obj):
                return obj

            while obj._getParent() is not None:
                if cond(obj):
                    break
                obj = obj._getParent()
            return obj

    # accept a visitor pattern
    @abc.abstractmethod
    def accept(self, visitor):
        """
        This can probably be removed if it doesn't do anything, only GangaObjet should inherit from Node
        """
        pass

    # mark object as "dirty" and inform the registry about it
    # the registry is always associated with the root object
    def _setDirty(self):
        """ Set the dirty flag all the way up to the parent"""
        self._dirty = True
        parent = self._getParent()
        if parent is not None:
            parent._setDirty()

    def _setFlushed(self):
        """
        This returns whether this object has been marked as dirty or not
        """
        self._dirty = False

    def printTree(self, f=None, sel=''):
        """
        This method will print a full tree which contains a full description of the class which of interest
        Args:
            f (stream): file-like output stream
            sel (str): Selection to display 
        """
        from GangaCore.GPIDev.Base.VPrinter import VPrinter
        self.accept(VPrinter(f, sel))

    def printSummaryTree(self, level=0, verbosity_level=0, whitespace_marker='', out=None, selection='', interactive=False):
        """If this method is overridden, the following should be noted:

        Args:
            level (int): the hierachy level we are currently at in the object tree.
            verbosity_level (int): How verbose the print should be. Currently this is always 0.
            whitespace_marker (str): If printing on multiple lines, this allows the default indentation to be replicated.
                               The first line should never use this, as the substitution is 'name = %s' % printSummaryTree()
            out (stream): An output stream to print to. The last line of output should be printed without a newline.'
            selection (str): See VPrinter for an explaintion of this.
            interactive (bool): Is this being printed to the interactive prompt
        """
        from GangaCore.GPIDev.Base.VPrinter import VSummaryPrinter
        self.accept(VSummaryPrinter(level, verbosity_level, whitespace_marker, out, selection, interactive))

    @abc.abstractmethod
    def __eq__(self, node):
        """
        This can probably be removed if it doesn't do anything, only GangaObjet should inherit from Node
        """
        pass

    def __ne__(self, node):
        """
        This can probably be removed if it doesn't do anything, only GangaObjet should inherit from Node
        """
        return not self.__eq__(node)

##########################################################################


def synchronised_get_descriptor(get_function):
    """
    This decorator should only be used on ``__get__`` method of the ``Descriptor``.
    Args:
        get_function (function): Function we intend to wrap with the soft/read lock
    """
    @functools.wraps(get_function)
    def get_decorator(self, obj, type_or_value):
        if obj is None:
            return get_function(self, obj, type_or_value)

        with obj._getRoot()._lock:
            return get_function(self, obj, type_or_value)

    return get_decorator


def synchronised_set_descriptor(set_function):
    """
    This decorator should only be used on ``__set__`` method of the ``Descriptor``.
    Args:
        set_function (function): Function we intend to wrap with the hard/write lock
    """
    def set_decorator(self, obj, type_or_value):
        root_obj = obj._getRoot()
        if obj is None:
            return set_function(self, obj, type_or_value, root_obj)

        with root_obj._lock:
            return set_function(self, obj, type_or_value)

    return set_decorator


class Descriptor(object):

    """
    This is the Descriptor class used to deal with object assignment ot attribtues of the GangaObject.
    This class handles the lazy-loading of an object from disk when needed to return a value stored in the Registry
    This class also handles thread-locking of a class including both the getter and setter methods to ensure object consistency
    """

    __slots__ = ('_name', '_item', '_checkset_name', '_filter_name', '_getter_name', '_checkset_name')

    def __init__(self, name, item):
        """
        Lets build a descriptor for this item with this name
        Args:
            name (str): Name of the attribute being wrapped
            item (Item): The Schema entry describing this attribute (Not currently used atm)
        """
        super(Descriptor, self).__init__()

        self._name = name
        self._item = item
        self._checkset_name = None
        self._filter_name = None

        self._getter_name = item['getter']
        self._checkset_name = item['checkset']
        self._filter_name = item['filter']

    @staticmethod
    def _bind_method(obj, name):
        """
        Method which returns the value for a given attribute of a name
        Args:
            name (str): This is the name of the attribute of interest
        """
        if name is None:
            return None
        return getattr(obj, name)

    def _check_getter(self):
        """
        This attribute checks to see if a getter has been assigned to an attribute or not to avoid it's assignment
        """
        if self._getter_name:
            raise AttributeError('cannot modify or delete "%s" property (declared as "getter")' % _getName(self))

    @synchronised_get_descriptor
    def __get__(self, obj, cls):
        """
        Get method of Descriptor
        This wraps the object in question with a read-lock which ensures object onsistency across multiple threads
        Args:
            obj (GangaObject): This is the object which controls the attribute of interest
            cls (class): This is the class of the Ganga Object which is being called
        """
        name = _getName(self)

        # If obj is None then the getter was called on the class so return the Item
        if obj is None:
            return cls._schema[name]

        if self._getter_name:
            # Fixme set the parent of Node objects?!?!
            return self._bind_method(obj, self._getter_name)()

        # First we want to try to get the information without prompting a load from disk

        # ._data takes priority ALWAYS over ._index_cache
        # This access should not cause the object to be loaded
        obj_data = obj._data
        try:
            return obj_data[name]
        except KeyError:
            pass

        # Then try to get it from the index cache
        obj_index = obj._index_cache
        try:
            return obj_index[name]
        except KeyError:
            pass

        # Since we couldn't find the information in the cache, we will need to fully load the object
        obj._loadObject()

        # Do we have the attribute now?
        try:
            return obj._data[name]
        except KeyError:
            pass

        # Last option: get the default value from the schema
        if obj._schema.hasItem(name):
            return obj._schema.getDefaultValue(name)

        raise AttributeError('Could not find attribute {0} in {1}'.format(name, obj))

    @staticmethod
    def cloneObject(v, obj, name):
        """
        Clone v using knowledge of the obj the attr is being set on and the name of self is the attribute name
        return a new instance of v equal to v
        Args:
            v (unknown): This is the object we want to clone
            obj (GangaObject): This is the parent object of the attribute
            name (str): This is the name of the attribute we want to assign the value of v to
        """
        item = obj._schema[name]

        if v is None:
            if item.hasProperty('category'):
                assertion = item['optional'] and (item['category'] != 'internal')
            else:
                assertion = item['optional']
            #assert(assertion)
            if assertion is False:
                logger.warning("Item: '%s'. of class type: '%s'. Has a Default value of 'None' but is NOT optional!!!" % (name, type(obj)))
                logger.warning("Please contact the developers and make sure this is updated!")
            return None
        elif isinstance(v, str):
            return str(v)
        elif isinstance(v, int):
            return int(v)
        elif isinstance(v, dict):
            new_dict = {}
            for key, item in v.items():
                new_dict[key] = Descriptor.cloneObject(item, obj, name)
            return new_dict
        else:
            if not isinstance(v, Node) and isinstance(v, (list, tuple)):
                try:
                    # must import here as will fail at the top
                    from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
                    new_v = GangaList()
                except ImportError:
                    new_v = []
                for elem in v:
                    new_v.append(Descriptor.cloneObject(elem, obj, name))
                #return new_v
            elif not isinstance(v, Node):
                if isclass(v):
                    new_v = v()
                else:
                    new_v = v
                if not isinstance(new_v, Node):
                    logger.error("v: %s" % v)
                    raise GangaException("Error: found Object: %s of type: %s expected an object inheriting from Node!" % (v, type(v)))
                else:
                    new_v = Descriptor.cloneNodeObject(new_v, obj, name)
            else:
                new_v = Descriptor.cloneNodeObject(v, obj, name)

            return new_v

    @staticmethod
    def cloneNodeObject(v, obj, name):
        """This copies objects inherited from Node class
        This is a call to deepcopy after testing to see if the object can be copied to the attribute
        Args:
            v (GangaObject): This is the value which we want to copy from
            obj (GangaObject): This is the object which controls the attribute we want to assign
            name (str): This is th name of the attribute which we're setting
        """

        item = obj._schema[name]
        if isinstance(v, GangaList):
            categories = v.getCategory()
            len_cat = len(categories)
            if (len_cat > 1) or ((len_cat == 1) and (categories[0] != item['category'])) and item['category'] != 'internal':
                # we pass on empty lists, as the catagory is yet to be defined
                from GangaCore.GPIDev.Base.Proxy import GangaAttributeError
                raise GangaAttributeError('%s: attempt to assign a list containing incompatible objects %s to the property in category "%s"' % (name, _getName(v), item['category']))
        else:
            if v._category not in [item['category'], 'internal'] and item['category'] != 'internal':
                from GangaCore.GPIDev.Base.Proxy import GangaAttributeError
                raise GangaAttributeError('%s: attempt to assign an incompatible object %s to the property in category "%s found cat: %s"' % (name, _getName(v), item['category'], v._category))

        v_copy = deepcopy(v)

        return v_copy

    @synchronised_set_descriptor
    def __set__(self, obj, val, root_obj=None):
        """
        Set method
        This wraps the given object with a lock preventing both read+write until this transaction is complete for consistency
        Args:
            self: attribute being changed or GangaCore.GPIDev.Base.Objects.Descriptor in which case _getName(self) gives the name of the attribute being changed
            obj (GanagObject): parent class which 'owns' the attribute
            val (unknown): value of the attribute which we're about to set
            root_obj (GangaObject): a pointer to the root object of obj
        """

        _set_name = _getName(self)

        if isinstance(val, str):
            if val:
                from GangaCore.GPIDev.Base.Proxy import stripProxy, runtimeEvalString
                val = stripProxy(runtimeEvalString(obj, _set_name, val))

        if hasattr(obj, '_checkset_name'):
            checkSet = self._bind_method(obj, self._checkset_name)
            if checkSet is not None:
                checkSet(val)

        if hasattr(obj, '_filter_name'):
            this_filter = self._bind_method(obj, self._filter_name)
            if this_filter:
                val = this_filter(val)

        self._check_getter()

        # make sure we have the session lock
        obj._getSessionLock(root_obj)

        # make sure the object is loaded if it's attached to a registry
        obj._loadObject()

        basic=False
        for i in [int, str, bool, type]:
            if isinstance(val, i):
                new_value = deepcopy(val)
                basic=True
                break
        if not basic:
            new_value = Descriptor.cleanValue(obj, val, _set_name)

        obj.setSchemaAttribute(_set_name, new_value)

        obj._setDirty()

    @staticmethod
    def cleanValue(obj, val, name):
        """
        This returns a new instance of the value which has been correctly copied if needed so we can assign it to the attribute of the class
        Args:
            obj (GangaObject): This is the parent object of the attribute which is being set
            val (unknown): This is the value we want to assign to the attribute
            name (str): This is the name of the attribute which we're changing
        """

        item = obj._schema[name]

        ## If the item has been defined as a sequence great, let's continue!
        if item['sequence']:
            # These objects are lists
            _preparable = True if item['preparable'] else False
            if val is None:
                new_val = None
            elif len(val) == 0:
                new_val = GangaList()
            else:
                if isinstance(item, ComponentItem):
                    new_val = makeGangaList(val, Descriptor.cloneListOrObject, obj, _preparable, (name, obj))
                else:
                    new_val = makeGangaList(val, None, obj, _preparable)
        else:
            ## Else we need to work out what we've got.
            if isinstance(item, ComponentItem):
                if isinstance(val, (list, tuple, GangaList)):
                    ## Can't have a GangaList inside a GangaList easily so lets not
                    if isinstance(obj, GangaList):
                        new_val = []
                    else:
                        new_val = GangaList()
                    # Still don't know if val list of object here
                    Descriptor.createNewList(new_val, val, Descriptor.cloneListOrObject, (name, obj))
                else:
                    # Still don't know if val list of object here
                    new_val = Descriptor.cloneListOrObject(val, (name, obj))
            else:
                new_val = val
                pass

        if isinstance(new_val, Node) and new_val._getParent() is not obj:
            new_val._setParent(obj)

        return new_val

    @staticmethod
    def cloneListOrObject(v, extra_args):
        """
        This clones the value v by determining if the value of v is a list or not.
        If v is a list then a new list is returned with elements copied via cloneObject
        if v is not a list then a new instance of the list is copied via cloneObject
        Args:
            v (unknown): Object we want a new copy of
            extra_args (tuple): Contains the name of the attribute being copied and the object which owns the object being copied
        """

        name=extra_args[0]
        obj=extra_args[1]
        if isinstance(v, (list, tuple, GangaList)):
            new_v = GangaList()
            for elem in v:
                new_v.append(Descriptor.cloneObject(elem, obj, name))
            return new_v
        else:
            return Descriptor.cloneObject(v, obj, name)

    def __delete__(self, obj):
        """
        Delete an attribute from the Descriptor(?) and Node
        Args:
            obj (GangaObject): This is the object which wants to have an attribute removed from it
        """
        del obj._data[_getName(self)]

    @staticmethod
    def createNewList(_final_list, _input_elements, action=None, extra_args=None):
        """ Create a new list object which contains the old object with a possible action parsing the elements before they're added
        Args:
            _final_list (list): The list object which is to have the new elements appended to it
            _input_elements (list): This is a list of the objects which are to be used to create new objects in _final_list
            action (function): A function which is to be called to create new objects for a list
            extra_args (tuple): Contains the name of the attribute being copied and the object which owns the object being copied
        """

        if extra_args:
            new_action = partial(action, extra_args=extra_args)

        if action is not None:
            for element in _input_elements:
                if extra_args:
                    _final_list.append(new_action(element))
                else:
                    _final_list.append(action(element))
        else:
            for element in _input_elements:
                _final_list.append(element)


class ObjectMetaclass(abc.ABCMeta):
    """
    This is a MetaClass...
    TODO explain what this does"""

    __slots__ = list()

    def __init__(cls, name, bases, this_dict):
        """
        Init method for a class of name, name
        TODO, explain what bases and this_dict are used for
        """

        super(ObjectMetaclass, cls).__init__(name, bases, this_dict)

        # all Ganga classes must have (even empty) schema
        if cls._schema is None:
            cls._schema = Schema(None, None)

        this_schema = cls._schema

        # This reduces the memory footprint
        # TODO explore migrating the _data_dict object to a non-dictionary type as it's a fixed size and we can potentially save big here!
        # Adding the __dict__ here is an acknowledgement that we don't control all Ganga classes higher up.
        cls.__slots__ = ('_index_cache_dict', '_registry', '_data_dict', '__dict__', '_proxyObject')

        # Add all class members of type `Schema.Item` to the _schema object
        # TODO: We _could_ add base class's Items here by going through `bases` as well.
        # We can't just yet because at this point the base class' Item has been overwritten with a Descriptor
        for member_name, member in this_dict.items():
            if isinstance(member, Item):
                this_schema.datadict[member_name] = member

        # sanity checks for schema...
        if '_schema' not in this_dict:
            s = "Class %s must _schema (it cannot be silently inherited)" % (name,)
            logger.error(s)
            raise GangaValueError(s)

        attrs_to_add = [ attr for attr, item in this_schema.allItems()]

        if hasattr(cls, '_additional_slots'):
            attrs_to_add += [_ for _ in cls._additional_slots]

        cls.__slots__ = ('_index_cache_dict', '_registry', '_data_dict', '__dict__', '_proxyObject') + tuple(attrs_to_add)

        # If a class has not specified a '_name' then default to using the class '__name__'
        if not cls.__dict__.get('_name'):
            cls._name = name

        if this_schema._pluginclass is not None:
            logger.warning('Possible schema clash in class %s between %s and %s', name, _getName(cls), _getName(this_schema._pluginclass))

        # export visible properties... do not export hidden properties
        # This constructs one Descriptor for each attribute which can be set for this class
        for attr, item in this_schema.allItems():
            setattr(cls, attr, Descriptor(attr, item))

        # additional check of type
        # bugfix #40220: Ensure that default values satisfy the declared types
        # in the schema
        for attr, item in this_schema.simpleItems():
            if not item['getter']:
                item._check_type(item['defvalue'], '.'.join([name, attr]), False)

        # create reference in schema to the pluginclass
        this_schema._pluginclass = cls

        # if we've not even declared this we don't want to use it!
        if not cls._declared_property('hidden') or cls._declared_property('enable_plugin'):
            allPlugins.add(cls, cls._category, _getName(cls))

        # create a configuration unit for default values of object properties
        if not cls._declared_property('hidden') or cls._declared_property('enable_config'):
            this_schema.createDefaultConfig()


@call_counter
@cpu_profiler
@mem_profiler
class GangaObject(Node, metaclass=ObjectMetaclass):
    _schema = None  # obligatory, specified in the derived classes
    _category = None  # obligatory, specified in the derived classes
    _exportmethods = []  # optional, specified in the derived classes

    # by default classes are not hidden, config generation and plugin
    # registration is enabled
    # optional, specify in the class if you do not want to export it publicly
    # in GPI,
    _hidden = 1
    # the class will not be registered as a plugin unless _enable_plugin is defined
    # the test if the class is hidden is performed by x._declared_property('hidden')
    # which makes sure that _hidden must be *explicitly* declared, not
    # inherited

    # additional properties that may be set in derived classes which were declared as _hidden:
    #   _enable_plugin = 1 -> allow registration of _hidden classes in the allPlugins dictionary
    # _enable_config = 1 -> allow generation of [default_X] configuration
    # section with schema properties

    # This boolean is used to stop a GangaObject from populating it's ._data dict with
    # default entries as evaluated from the schema.
    # This is useful when reading many objects from disk as this wastes CPU as the entry is being overridden
    _should_init = True
    _should_load = False

    @classmethod
    def getNew(cls, should_load=False, should_init=False):
        """
        Returns a new instance of this class type without a populated Schema.
        This should be an object which has all of the core logic initialized correctly.
        Args:
            should_load (bool): Should the class __init__ method be called by this method or should we return the base object?
        """
        # Build an object of this type
        returnable = cls.__new__(cls, (), {})

        # We have implicit knowledge about whether we should attempt to load this object from disk
        setattr(returnable, '_should_load', should_load)

        Node.__init__(returnable)

        setattr(returnable, '_should_init', should_init)
        GangaObject.__init__(returnable)
        try:
            # Initialize the most derived class to get all of the goodness needed higher up.
            returnable.__class__.__init__(returnable)
        except:
            logger.debug("Broken init method for class: %s trying to proceed silently" % cls.__name__)
        setattr(returnable, '_should_init', False)

        # Return the newly initialized object
        return returnable

    # must be fully initialized
    def __init__(self):
        """
        Main GangaObject that many classes inherit from
        """
        super(GangaObject, self).__init__(None)

        self._index_cache_dict = {}
        self._registry = None
        self._data_dict = {}

        # Just a flag to prevent expensive double-declarations and to avoid this where needed
        if self._should_init:
            self.populate_from_schema()

    def populate_from_schema(self):
        """
        Populate the data dict from the schema defaults
        """
        self._data_dict = dict.fromkeys(self._schema.datadict)
        for attr, item in self._schema.allItems():
            ## If an object is hidden behind a getter method we can't assign a parent or defvalue so don't bother - rcurrie
            if item.getProperties()['getter'] is None:
                setattr(self, attr, self._schema.getDefaultValue(attr))

    @synchronised
    def accept(self, visitor):
        """
        This accepts a VPrinter or VStreamer class instance visitor which is used to display the contents of the object according the Schema
        Args:
            visitor (VPrinter, VStreamer): An instance which will produce a display of this contents of this object
        """

        if not hasattr(self, '_schema'):
            return
        elif self._schema is None:
            visitor.nodeBegin(self)
            visitor.nodeEnd(self)
            return

        visitor.nodeBegin(self)

        for (name, item) in self._schema.simpleItems():
            if item['visitable']:
                visitor.simpleAttribute(self, name, getattr(self, name), item['sequence'])

        for (name, item) in self._schema.sharedItems():
            if item['visitable']:
                visitor.sharedAttribute(self, name, getattr(self, name), item['sequence'])

        for (name, item) in self._schema.componentItems():
            if item['visitable']:
                visitor.componentAttribute(self, name, getattr(self, name), item['sequence'])

        visitor.nodeEnd(self)

    def copyFrom(self, srcobj, _ignore_atts=None):
        # type: (GangaObject, Optional[Sequence[str]]) -> None
        """
        copy all the properties recursively from the srcobj
        if schema of self and srcobj are not compatible raises a ValueError
        ON FAILURE LEAVES SELF IN INCONSISTENT STATE
        Args:
            srcobj (GangaObject): This is the ganga object which is to have it's contents from
            _ignore_atts (list): This is a list of attribute names which are to not be copied
        """

        if _ignore_atts is None:
            _ignore_atts = []
        _srcobj = srcobj
        # Check if this object is derived from the source object, then the copy
        # will not throw away information

        if not hasattr(_srcobj, '__class__') and not inspect.isclass(_srcobj.__class__):
            raise GangaValueError("Can't copyFrom a non-class object: %s isclass: %s" % (_srcobj, inspect.isclass(_srcobj)))

        if not isinstance(self, _srcobj.__class__) and not isinstance(_srcobj, self.__class__):
            raise GangaValueError("copyFrom: Cannot copy from %s to %s!" % (_getName(_srcobj), _getName(self)))

        if not hasattr(self, '_schema'):
            logger.debug("No Schema found for myself")
            return

        if self._schema is None and _srcobj._schema is None:
            logger.debug("Schema object for one of these classes is None!")
            return

        if _srcobj._schema is None:
            self._schema = None
            return

        self._actually_copyFrom(_srcobj, _ignore_atts)

        ## Fix some objects losing parent knowledge
        src_dict = srcobj.__dict__
        for key, val in src_dict.items():
            this_attr = getattr(srcobj, key)
            if isinstance(this_attr, Node) and key not in do_not_copy:
                #logger.debug("k: %s  Parent: %s" % (key, (srcobj)))
                if this_attr._getParent() is not srcobj:
                    this_attr._setParent(srcobj)

    def _actually_copyFrom(self, _srcobj, _ignore_atts):
        # type: (GangaObject, Optional[Sequence[str]]) -> None

        for name, item in self._schema.allItems():
            if name in _ignore_atts:
                continue

            #logger.debug("Copying: %s : %s" % (name, item))
            if name == 'application' and hasattr(_srcobj.application, 'is_prepared'):
                _app = _srcobj.application
                if _app.is_prepared not in [None, True]:
                    _app.incrementShareCounter(_app.is_prepared)

            if not self._schema.hasAttribute(name):
                #raise ValueError('copyFrom: incompatible schema: source=%s destination=%s'%(_getName(_srcobj), _getName(self)))
                if not hasattr(self, name):
                    setattr(self, name, self._schema.getDefaultValue(name))
                this_attr = getattr(self, name)
                if isinstance(this_attr, Node) and name not in do_not_copy:
                    if this_attr._getParent() is not self:
                        this_attr._setParent(self)
            elif not item['copyable']: ## Default of '1' instead of True...
                if not hasattr(self, name):
                    setattr(self, name, self._schema.getDefaultValue(name))
                this_attr = getattr(self, name)
                if isinstance(this_attr, Node) and name not in do_not_copy:
                    if this_attr._getParent() is not self:
                        this_attr._setParent(self)
            else:
                copy_obj = deepcopy(getattr(_srcobj, name))
                setattr(self, name, copy_obj)

    def __eq__(self, obj):
        """
        Compare this object to an other object obj
        Args:
            obj (unknown): Compare which self is to be compared to
        """
        if self is obj:
            return True

        if not isinstance(obj, type(self)):
            return False

        # Compare the schemas against each other
        if self._schema is None and obj._schema is None:
            return True  # If they're both `None`
        elif self._schema is None or obj._schema is None:
            return False  # If just one of them is `None`
        elif not self._schema.isEqual(obj._schema):
            return False  # Both have _schema but do not match

        # Check each schema item in turn and check for equality
        for (name, item) in self._schema.allItems():
            if item['comparable']:
                #logger.info("testing: %s::%s" % (_getName(self), name))
                if getattr(self, name) != getattr(obj, name):
                    #logger.info( "diff: %s::%s" % (_getName(self), name))
                    return False

        return True

    @property
    def _data(self):
        """
        This returns the internal data dictionary of this class
        This should be treated as strictly internal and shouldn't be modified outside of GangaCore.GPIDev.Base or GangaCore.Core.GangaRegistry
        """
        # type: () -> Dict[str, Any]
        return self._data_dict

    @_data.setter
    def _data(self, new_data):
        """
        Set the ._data attribute correctly for this object
        This should be treated as strictly internal and shouldn't be modified outside of GangaCore.GPIDev.Base or GangaCore.Core.GangaRegistry
        Args:
            new_data (dict): This is the external dictionary which is to be assigned to the internal dictionary
        """
        # type: (Dict[str, Any]) -> None
        for v in new_data.values():
            if isinstance(v, Node) and v._getParent() is not self:
                v._setParent(self)
        self._data_dict = new_data

    def setSchemaAttribute(self, attrib_name, attrib_value):
        # type: (str, Any) -> None
        """
        This sets the value of a schema attribute directly by circumventing the descriptor

        Args:
            attrib_name (str): the name of the schema attribute
            attrib_value (unknown): the value to set it to
        """
        self._data[attrib_name] = attrib_value
        if isinstance(attrib_value, Node) and attrib_value._getParent() is not self:
            self._data[attrib_name]._setParent(self)

    @property
    def _index_cache(self):
        """
        This returns the index cache.
        If this object has been fully loaded into memory and has a Registry assoicated with it it's created dynamically
        If this object isn't fully loaded or has no Registry associated with it the index_cache should be returned whatever it is
        """
        if self._fullyLoadedFromDisk():
            if self._getRegistry() is not None:
                # Fully loaded so lets regenerate this on the fly to avoid losing data
                return self._getRegistry().getIndexCache(self)
            else:
                # No registry therefore can't work out the Cache, probably empty, lets return that
                return self._index_cache_dict
        # Not in registry or not loaded, so can't re-generate if requested
        return self._index_cache_dict

    @_index_cache.setter
    def _index_cache(self, new_index_cache):
        """
        Set the index cache to have some new value which is defined externally.
        (This should __NEVER__ be done on live in-memory objects
        Args:
            new_inde_cache (dict): Dict of new entries for the new_index_cache
        """
        if self._fullyLoadedFromDisk():
            logger.debug("Warning: Setting IndexCache data on live object, please avoid!")
        self._index_cache_dict = new_index_cache

    def _fullyLoadedFromDisk(self):
        # type: () -> bool
        """This returns a boolean. and it's related to if self has_loaded in the Registry of this object"""
        if self._getRegistry() is not None:
            return self._getRegistry().has_loaded(self)
        return True

    @staticmethod
    def __incrementShareRef(obj, attr_name):
        """
        This increments the shareRef of the prep registry according to the attr_name.name
        Args:
            obj (GangaObject): This is the object which has the attr_name attribute
            attr_name (str): This is the name of the attribute which contains a SharedDir
        """
        shared_dir = getattr(obj, attr_name)

        if hasattr(shared_dir, 'name'):

            from GangaCore.Core.GangaRepository import getRegistry
            shareref = getRegistry("prep").getShareRef()

            logger.debug("Increasing shareref")
            shareref.increase(shared_dir)

    def __copy__(self):
        obj = self.getNew()
        # FIXME: this is different than for deepcopy... is this really correct?
        this_dict = copy(self.__dict__)
        for elem in this_dict.keys():
            if elem not in do_not_copy:
                this_dict[elem] = copy(this_dict[elem])
            else:
                this_dict[elem] = None
        obj._setParent(self._getParent())
        obj._index_cache = {}
        obj._registry = self._registry
        return obj

    # on the deepcopy reset all non-copyable properties as defined in the
    # schema
    def __deepcopy__(self, memo=None):
        """
        Perform a deep copy of the GangaObject class
        Args:
            memo (unknown): Used to track infinite loops etc in the deep-copy of objects
        """

        true_parent = self._getParent()

        self_copy = self.getNew()

        global do_not_copy
        if self._schema is not None:
            for name, item in self._schema.allItems():
                if not item['copyable'] or name in do_not_copy or not hasattr(self, name):
                    setattr(self_copy, name, self._schema.getDefaultValue(name))
                else:
                    setattr(self_copy, name, deepcopy(getattr(self, name)))

                this_attr = getattr(self_copy, name)
                if isinstance(this_attr, Node) and this_attr._getParent() is not self_copy:
                    this_attr._setParent(self_copy)

                if item.isA(SharedItem):
                    self.__incrementShareRef(self_copy, name)

        for k, v in self.__dict__.items():
            if k not in do_not_copy:
                try:
                    self_copy.__dict__[k] = deepcopy(v)
                except:
                    self_copy.__dict__[k] = v

        if true_parent is not None:
            if self._getParent() is not true_parent:
                self._setParent(true_parent)
            if self_copy._getParent() is not true_parent:
                self_copy._setParent(true_parent)
        setattr(self_copy, '_registry', self._registry)
        return self_copy

    def clone(self):
        """Clone self and return a properly initialized object"""
        return deepcopy(self)

    def _getIOTimeOut(self):
        """
        Get the DiskIOTimeout or 5 if this is not defined in the config
        """
        from GangaCore.Utility.Config.Config import getConfig, ConfigError
        try:
            _timeOut = getConfig('Configuration')['DiskIOTimeout']
        except ConfigError as err:
            _timeOut = 5. # 5sec hardcoded default
        return _timeOut

    def _getSessionLock(self, root=None):
        """Acquires the session lock on this object"""
        if root:
            r=root
        else:
            r = self._getRoot()
        reg = r._registry
        if reg is not None:
            reg._acquire_session_lock(r)

    def _releaseSessionLockAndFlush(self):
        """ Releases the session lock for this object
        Please use only if the object is expected to be used by other sessions"""
        r = self._getRoot()
        reg = r._registry
        if reg is not None:
            reg._release_session_lock_and_flush(r)

    def _loadObject(self):
        """If there's an attached registry then ask it to load this object"""
        if self._should_load and self._registry is not None:
            self._registry._load(self)
            self._should_load = False

    # define when the object is read-only (for example a job is read-only in
    # the states other than new)
    def _readonly(self):
        """
        Returns a 1 or 0 depending on if this object is read-only
        TODO: make this True/False
        """
        r = self._getRoot()
        # is object a root for itself? check needed otherwise infinite
        # recursion
        if r is None or r is self:
            return 0
        else:
            return r._readonly()

    def _setRegistry(self, registry):
        """
        Set the Registry of the GangaObject which will manage it. This can only
        set the registry *to* or *from* None and never from one registry to
        another.
        Args:
            registry (GangaRegistry): This is a Ganga Registry object
        """
        assert self._getParent() is None, 'Can only set the registry of a root object'
        if registry is None or self._registry is None:
            self._registry = registry
        elif registry is not self._registry:
            raise RuntimeError('Cannot set registry of {0} to {1} if one is already set ({2}).'.format(type(self), registry, self._registry))

    # get the registry for the object by getting the registry associated with
    # the root object (if any)
    def _getRegistry(self):
        """
        Get the registry which is managing this GangaObject
        The registry is only managing a root object so it gets this first
        """
        return self._getRoot()._registry

    def _getRegistryID(self):
        """
        Get the ID of self within a Registry
        This is normally the .id of an object itself but there is no need for it to be implemented this way
        """
        try:
            return self._registry.find(self)
        except AttributeError as err:
            logger.debug("_getRegistryID Exception: %s" % err)
            return None

    def _setFlushed(self, auto_load_deps=True):
        """
        Un-Set the dirty flag all of the way down the schema.
        Args:
            auto_load_deps (bool): Should we attempt to get objects which may be unloaded and load them via the schema?
        """
        if self._schema and auto_load_deps:
            for k in self._schema.allItemNames():
                this_attr = getattr(self, k)
                if isinstance(this_attr, Node):
                    if not this_attr._dirty:
                        continue
                else:
                    continue
                ## Avoid attributes the likes of job.master which crawl back up the tree
                k_props = self._schema[k].getProperties()
                if not k_props['visitable'] or k_props['transient']:
                    continue
                if isinstance(this_attr, Node) and this_attr._dirty:
                    this_attr._setFlushed()
        super(GangaObject, self)._setFlushed()

    # post __init__ hook automatically called by GPI Proxy __init__
    def _auto__init__(self):
        """
        This is called when an object is constructed from infront of the Proxy automatically, or manually when mimicing the behavior of the IPython prompt
        default behavior is to do nothing
        """
        pass

    # return True if _name attribute was explicitly defined in the class
    # this means that implicit (inherited) _name attribute has no effect in the derived class
    # example: cls._declared_property('hidden') => True if there is class
    # attribute _hidden explicitly declared
    @classmethod
    def _declared_property(cls, name):
        return '_' + name in cls.__dict__

    # get the job object associated with self or raise an assertion error
    # the FIRST PARENT Job is returned...
    # this method is for convenience and may well be moved to some subclass
    def getJobObject(self):
        """
        Return the parent Job which manages this object or throw an AssertionError is non exists
        Unknown: Why is this a GangaObject method and not a Node method?
        """
        from GangaCore.GPIDev.Lib.Job import Job
        if self._getParent() is not None:
            r = self._getRoot(cond=lambda o: isinstance(o, Job))
            if not isinstance(r, Job):
                raise AssertionError('No Job associated with object instead root=\'%s\' for \'%s\'' % (repr(r), type(r)))
            return r
        elif isinstance(self, Job):
            return self
        raise AssertionError('No Parent associated with object \'%s\'' % repr(self))

    # Customization of the GPI attribute assignment: Attribute Filters
    #
    # Example of usage:
    # if some properties depend on the value of other properties in a complex way such as:
    # changing platform of Gaudi should change the version if it is not supported... etc.
    #
    # Semantics:
    #  gpi_proxy.x = v        --> stripProxy(gpi_proxy)._attribute_filter__set__('x',v)
    #  gpi_proxy.y = [v1,v2]  --> stripProxy(gpi_proxy)._attribute_filter__set__('x',[v1,v2])
    #
    #  is used for *all* kinds of attributes (component and simple)
    #
    # Attribute Filters are used mainly for side effects such as modification of the state of the self object
    # (as described above).
    #
    # Attribute Filter may also perform an additional conversion of the value which is being assigned.
    #
    # Returns the attribute value (converted or original)
    #
    def _attribute_filter__set__(self, name, v):
        if hasattr(v, '_on_attribute__set__'):
            return v._on_attribute__set__(self, name)
        return v


# define the default component object filter:
# obj.x = "Y"   <=> obj.x = Y()


def string_type_shortcut_filter(val, item):
    """
    Filter which allows for "obj.x = "Y"   <=> obj.x = Y()"
    TODO evaluate removing this and the architecture behind it
    """
    if isinstance(val, type('')):
        if item is None:
            raise GangaValueError('cannot apply default string conversion, probably you are trying to use it in the constructor')
        from GangaCore.Utility.Plugin import allPlugins, PluginManagerError
        try:
            obj = allPlugins.find(item['category'], val)()
            obj._auto__init__()
            return obj
        except PluginManagerError as err:
            logger.debug("string_type_shortcut_filter Exception: %s" % err)
            raise GangaValueError('Cannot assign string to object, are you sure this is correct?\n%s' % err)
    return None

# FIXME: change into classmethod (do they inherit?) and then change stripComponentObject to use class instead of
# FIXME: object (object model clearly fails with sequence of Files)
# FIXME: test: ../bin/ganga -c local_lhcb.ini run.py
# TestNativeSpecific.testFileSequence


from .Filters import allComponentFilters
allComponentFilters.setDefault(string_type_shortcut_filter)

from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList, makeGangaList

#
#
# $Log: not supported by cvs2svn $
# Revision 1.5.2.9  2009/07/14 14:44:17  ebke
# * several bugfixes
# * changed indexing for XML/Pickle
# * introduce index update minimal time of 20 seconds (reduces lag for typing 'jobs')
# * subjob splitting and individual flushing for XML/Pickle
#
# Revision 1.5.2.8  2009/07/13 22:10:53  ebke
# Update for the new GangaRepository:
# * Moved dict interface from Repository to Registry
# * Clearly specified Exceptions to be raised by Repository
# * proper exception handling in Registry
# * moved _writable to _getWriteAccess, introduce _getReadAccess
# * clarified locking, logic in Registry, less in Repository
# * index reading support in XML (no writing, though..)
# * general index reading on registry.keys()
#
# Revision 1.5.2.7  2009/07/10 12:14:10  ebke
# Fixed wrong sequence in __set__: only dirty _after_ writing!
#
# Revision 1.5.2.6  2009/07/10 11:33:06  ebke
# Preparations and fixes for lazy loading
#
# Revision 1.5.2.5  2009/07/08 15:27:50  ebke
# Removed load speed bottleneck for pickle - reduced __setstate__ time by factor 3.
#
# Revision 1.5.2.4  2009/07/08 12:51:52  ebke
# Fixes some bugs introduced in the latest version
#
# Revision 1.5.2.3  2009/07/08 12:36:54  ebke
# Simplified _writable
#
# Revision 1.5.2.2  2009/07/08 11:18:21  ebke
# Initial commit of all - mostly small - modifications due to the new GangaRepository.
# No interface visible to the user is changed
#
# Revision 1.5.2.1  2009/06/04 12:00:37  moscicki
# *** empty log message ***
#
# Revision 1.5  2009/05/20 13:40:22  moscicki
# added filter property for GangaObjects
#
# added Utility.Config.expandgangasystemvars() filter which expands @{VAR} in strings, where VAR is a configuration option defined in the System section
# Usage example: specify @{GANGA_PYTHONPATH} in the configuration file to make pathnames relative to the location of Ganga release; specify @{GANGA_VERSION} to expand to current Ganga version. etc.
#
# modified credentials package (ICommandSet) to use the expandgangasystemvars() filter.
#
# Revision 1.4  2009/04/27 09:22:56  moscicki
# fix #29745: use __mro__ rather than first-generation of base classes
#
# Revision 1.3  2009/02/24 14:57:56  moscicki
# set parent correctly for GangaList items (in __setstate__)
#
# Revision 1.2  2008/09/09 14:37:16  moscicki
# bugfix #40220: Ensure that default values satisfy the declared types in the schema
#
# factored out type checking into schema module, fixed a number of wrongly declared schema items in the core
#
# Revision 1.1  2008/07/17 16:40:52  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.27.4.10  2008/03/31 15:30:26  kubam
# More flexible internal logic of hiding the plugin classes derived from GangaObject:
#
# # by default classes are not hidden, config generation and plugin registration is enabled
#
# _hidden       = 1  # optional, specify in the class if you do not want to export it publicly in GPI,
#                    # the class will not be registered as a plugin unless _enable_plugin is defined
#                    # the test if the class is hidden is performed by x._declared_property('hidden')
#
# # additional properties that may be set in derived classes which were declared as _hidden:
# #   _enable_plugin = 1 -> allow registration of _hidden classes in the allPlugins dictionary
# #   _enable_config = 1 -> allow generation of [default_X] configuration section with schema properties
#
# This fixes: bug #34470: config [defaults_GridProxy] missing (issue of _hidden property of GangaObject)
#
# Revision 1.27.4.9  2008/02/28 15:44:31  moscicki
# fixed set parent problem for GangaList (and removed Will's hack which was already commented out)
#
# Revision 1.27.4.8  2008/02/12 09:25:24  amuraru
# removed Will's hack to set the parent, it causes side effects in subjobs accessor(to be checked)
#
# Revision 1.27.4.7  2008/02/06 17:02:00  moscicki
# small fix
#
# Revision 1.27.4.6  2008/02/06 13:00:21  wreece
# Adds a bit of a HACK as a tempory measure. The parent of a GangaList is being lost somewhere (I suspect due to a copy). I've added it back manually in __get__.
#
# Revision 1.27.4.5  2008/02/06 09:28:48  wreece
# First pass at a cleanup of the gangalist stuff. I've made changes so the diffs with the 4.4 series are more transparent. Currently still test failures.
#
# Revision 1.27.4.4  2007/12/18 09:05:04  moscicki
# integrated typesystem from Alvin and made more uniform error reporting
#
# Revision 1.27.4.3  2007/11/14 13:03:54  wreece
# Changes to make shortcuts work correctly with gangalists. all but one tests should now pass.
#
# Revision 1.27.4.2  2007/11/07 15:10:02  moscicki
# merged in pretty print and GangaList support from ganga-5-dev-branch-4-4-1-will-print branch
#
#
# Revision 1.27.4.1  2007/10/30 15:25:53  moscicki
# fixed #29745: Inherited functions of GangaObjects cannot be exported via _exportmethods
#
# Revision 1.27.8.3  2007/10/30 21:22:02  wreece
# Numerous small fixes and work arounds. And a new test.
#
# Revision 1.27.8.2  2007/10/30 14:30:23  wreece
# Non-working update. Adds in Kuba's exported methods dodge. It is now possible to define a _export_ version of a method for external use and a undecorated method for internal use.
#
# Revision 1.27.8.1  2007/10/30 12:12:08  wreece
# First version of the new print_summary functionality. Lots of changes, but some known limitations. Will address in next version.
#
# Revision 1.27  2007/07/27 13:52:00  moscicki
# merger updates from Will (from GangaMergers-1-0 branch)
#
# Revision 1.26.12.1  2007/05/14 13:32:11  wreece
# Adds the merger related code on a seperate branch. All tests currently
# run successfully.
#
# Revision 1.26  2006/07/27 20:09:54  moscicki
# _getJobObject() renamed -> getJobObject() and finding FIRST PARENT JOB
# _getRoot() : cond optional argument (to cut the search path e.g. for getJobObject())
#
# getDefaultValue() moved to Schema object
#
# "checkset" metaproperty implemented in schema (to check the direct updates to job.status)
# modified the "getter" metaproperty implementation
# create automatically the configuration units with default property values
#
# Revision 1.25  2006/02/10 14:18:44  moscicki
# removed obsoleted eval for default properties from config file
#
# Revision 1.24  2006/01/09 16:36:55  moscicki
#  - support for defining default properties in the configuration
#    config file example:
#
#      [LSF_Properties]
#      queue = myQueue
#
#      [LCG_Properties]
#      requirements.CE = myCE
#
#      [Job_Properties]
#      application = DaVinci
#      application.version = v99r2
#
# Revision 1.23  2005/12/02 15:27:13  moscicki
# support for "visitable" metaproperty
# support for hidden GPI classes
# "hidden" properties not visible in the proxies
# support for 'getter' type of property (a binding) for job.master
#
# Revision 1.22  2005/11/14 10:29:55  moscicki
# adapted to changed interface of allPlugins.add() method
#
# Revision 1.21  2005/11/01 16:43:51  moscicki
# added isStringLike condition to avoid unnecesary looping over strings
#
# Revision 1.20  2005/11/01 14:04:55  moscicki
# added missing implementation of GangaObject.__setstate__
# fixed the implementation of Node.__setstate__ to correctly set parent for objects which are contained in iterable simple items (lists, dictionaries) etc. (contributed by AS)
#
# Revision 1.19  2005/11/01 11:09:48  moscicki
# changes to support import/export (KH):
#   - The printTree method accepts an additional argument, which
#     it passes on to VPrinter as the value for "selection".
#
# Revision 1.18  2005/08/26 09:53:24  moscicki
# copy __doc__ for exported method
# _getJobObject() helper method in GangaObject
#
# Revision 1.17  2005/08/24 15:41:19  moscicki
# automatically generated help for properties, disabled the SchemaHelper and few other improvements to the help system
#
# Revision 1.16  2005/08/23 17:15:06  moscicki
# *** empty log message ***
#
#
#



# This class (SessionLockManager) encapsulates fcntl-based locking that works on NFS, AFS and locally.
# You can use
# python SessionLock.py {1|2}
# to run locking tests (run several instances in the same directory, from
# different machines)



import functools
import threading
from collections import namedtuple

import os
import time
import errno
import fcntl
import random
import datetime
import getpass
from pipes import quote

import pickle as pickle

from GangaCore.Utility.logging import getLogger

from GangaCore.Utility.Config.Config import getConfig, ConfigError
from GangaCore.GPIDev.Base.Proxy import getName

from GangaCore.Core.GangaThread import GangaThread
from GangaCore.Core.GangaRepository import RepositoryError

logger = getLogger()

def lock_synch(f):
    """  
        This decorator must be attached to a method on a ``Node`` subclass
        It uses the object's lock to make sure that the object is held for the duration of the decorated function
        Args:
        f (function): This is the function which we want to wrap
    """  
    @functools.wraps(f)
    def decorated_lock(self, *args, **kwargs):
        with self.sync_lock:
            return f(self, *args, **kwargs)
    return decorated_lock

class FixedLockManager(object):

    """ Class with thread that keeps a single fixed lockfile which is removed on exit.
    This explicitly forbids multiple parallel Ganga sessions.
    """

    __slots__ = ('sync_lock', 'locked', 'repo', 'cntfn', 'global_lock', 'count')

    def __init__(self, repo, root, name, minimum_count=0):


        # Used for self runtime thread safety
        self.sync_lock = threading.RLock()

        # Required to pretend there are some locked jobs from other sessions
        self.locked = set()

        self.repo = repo
        self.mkdir(root)
        realpath = os.path.realpath(root)

        # Location of the count file for this repo
        self.cntfn = os.path.join(realpath, name, "cnt")

        # Required for lock files
        sessions_folder = os.path.join(realpath, "sessions")
        if not os.path.exists(sessions_folder):
            os.mkdir(sessions_folder) 

        # Location of fixed lock for this repo
        self.global_lock = os.path.join(sessions_folder, '%s_fixed_lock' % name)
        self.count = minimum_count
        if os.path.exists(self.global_lock):
            msg = "\n\nCannot start this registry: %s due to a lock file already existing: '%s'\n" % (name, self.global_lock)
            msg += "If you are trying to run only 1 ganga session please remove this file and re-launch ganga\n"
            raise RepositoryError(repo, msg)

        with open(self.global_lock, 'w'):
            pass

    @lock_synch
    def mkdir(self, dn):                                                                                                                                                                                           
        """Make sure the given directory exists"""
        try:
            if not os.path.exists(dn):
                os.makedirs(dn)
        except OSError as x:
            if x.errno != errno.EEXIST:
                raise RepositoryError(self.repo, "OSError on directory create: %s" % x)

    def startup(self):
        pass

    def finish_startup(self):
        pass

    def updateNow(self):
        pass

    def shutdown(self):
        """Shutdown the thread and locking system (on ganga shutdown or repo error)"""
        os.unlink(self.global_lock)
    
    def session_read(self, fn):
        return set()

    def session_write(self):
        pass

    @lock_synch
    def cnt_read(self):
        """Reads the counter file.
        """
        if not os.path.exists(self.cntfn):
            self.cnt_write()
        with open(self.cntfn) as fd:
            # 100 bytes should be enough for any ID. Can raise ValueErrorr
            _output = int(fd.readline())
        self.count = max(self.count, _output)
        return _output

    @lock_synch
    def cnt_write(self):
        """ Writes the counter to the counter file.
        """
        with open(self.cntfn, 'w') as fd:
            fd.write(str(self.count))

    @lock_synch
    def make_new_ids(self, n):
        """This bumps the registry count by n ids"""
        newcount = self.cnt_read()
        ids = list(range(newcount, newcount + n))
        self.count = ids[-1]+1
        self.cnt_write()
        return list(ids)

    def lock_ids(self, ids):
        """Pretend to lock IDs"""
        return list(ids)

    def release_ids(self, ids):
        """Pretend to release IDs"""
        return list(ids)

    def check(self):
        """ No need to check as we're disk/runtime consistent by construction. i.e. no parallel sessions """
        pass

    def get_lock_session(self, id):
        """Return a dummy str of self as only we can own a session
        """
        return "self"

    def get_other_sessions(self):
        """Return an empty list as No other sessions by definition
        """
        return []

    def reap_locks(self):
        """ Return True to respect checks on SessionLock"""
        return True




import os
import base64
import subprocess
import threading
import pickle as pickle
import signal
from copy import deepcopy
from GangaCore.Core.exceptions import GangaException
from GangaCore.Utility.logging import getLogger
logger = getLogger()


def bytes2string(obj):
    if isinstance(obj, bytes):
        return obj.decode("utf-8")
    if isinstance(obj, dict):
        return {bytes2string(key): bytes2string(value) for key, value in obj.items()}
    if isinstance(obj, list):
        return [bytes2string(item) for item in obj]
    if isinstance(obj, tuple):
        return tuple(bytes2string(item) for item in obj)
    return obj


def env_update_script(indent=''):
    """ This function creates an extension to a python script, or just a python script to be run at the end of the
    piece of code we're interested in.
    This piece of code will dump the environment after the execution has taken place into a temporary file.
    This returns a tuple of the script it's generated and the pipes file handlers used to store the end in memory
    Args:
        indent (str): This is the indent to apply to the script if this script is to be appended to a python file
    """
    fdread, fdwrite = os.pipe()
    os.set_inheritable(fdwrite, True)
    this_script = '''
import os
import pickle as pickle
with os.fdopen(###FD_WRITE###,'wb') as envpipe:
    pickle.dump(dict(os.environ), envpipe, 2)
'''
    from GangaCore.GPIDev.Lib.File.FileUtils import indentScript
    script = indentScript(this_script, '###INDENT###')

    script = script.replace('###INDENT###'  , indent      )\
                   .replace('###FD_READ###' , str(fdread) )\
                   .replace('###FD_WRITE###', str(fdwrite))

    return script, (fdread, fdwrite)


def python_wrapper(command, python_setup='', update_env=False, indent=''):
    """ This section of code wraps the given python command inside a small wrapper class to allow us to control the output.
    Optionally we can also append to the end of this file a script to allow us to extract the environment after we've
    finished executing our command.
    Args:
        command (str): This is the python code to be executed (can be multi-line)
        python_setup (str): This is some python code to be executed before the python code in question (aka a script header.
        update_env (bool): Contol whether we want to capture the env after running
        indent (str): This allows for an indent to be applied to the script so it can be placed inside other python scripts
    This returns the file handler objects for the env_update_script, the python wrapper itself and the script which has been generated to be run
    """
    fdread, fdwrite = os.pipe()
    os.set_inheritable(fdwrite, True)
    this_script = '''
from __future__ import print_function
import os, sys, traceback
import pickle as pickle
with os.fdopen(###PKL_FDWRITE###, 'wb') as PICKLE_STREAM:
    def output(data):
        pickle.dump(data, PICKLE_STREAM, 2)
    local_ns = {'pickle'        : pickle,
                'PICKLE_STREAM' : PICKLE_STREAM,
                'output'        : output}
    try:
        full_command = """###SETUP### """
        full_command += """ \n###COMMAND### """
        exec(full_command, local_ns)
    except:
        pickle.dump(traceback.format_exc(), PICKLE_STREAM, 2)

'''
    from GangaCore.GPIDev.Lib.File.FileUtils import indentScript
    script = indentScript(this_script, '###INDENT###')

    script = script.replace('###INDENT###'     , indent              )\
                   .replace('###SETUP###'      , python_setup.strip())\
                   .replace('###COMMAND###'    , command.strip()     )\
                   .replace('###PKL_FDREAD###' , str(fdread)         )\
                   .replace('###PKL_FDWRITE###', str(fdwrite)        )
    env_file_pipes = None
    if update_env:
        update_script, env_file_pipes = env_update_script()
        script += update_script
    return script, (fdread, fdwrite), env_file_pipes


def __reader(pipes, output_ns, output_var, require_output):
    """ This function un-pickles a pickle from a file and return it as an element in a dictionary
    Args:
        pipes (tuple): This is a tuple containing the (read_pipe, write_pipe) from os.pipes containing the pickled object
        output_ns (dict): This is the dictionary we should put the un-pickled object
        output_var (str): This is the key we should use to determine where to put the object in the output_ns
        require_output (bool): Should the reader give a warning if the pickle stream is not readable
    """
    os.close(pipes[1])
    with os.fdopen(pipes[0], 'rb') as read_file:
        try:
            # rcurrie this deepcopy hides a strange bug that the wrong dict is sometimes returned from here. Remove at your own risk
            output_ns[output_var] = deepcopy(pickle.load(read_file))
        except UnicodeDecodeError:
            output_ns[output_var] = deepcopy(bytes2string(pickle.load(read_file, encoding="bytes")))
        except Exception as err:
            if require_output:
                logger.error('Error getting output stream from command: %s', err)


def __timeout_func(process, timed_out):
    """ This function is used to kill functions which are timing out behind the scenes and taking longer than a
    threshold time to execute.
    Args:
        process (class): This is a subprocess class which knows of the pid of wrapping thread around the command we want to kill
        timed_out (Event): A threading event to be set when the command has timed out
    """
        
    if process.returncode is None:
        timed_out.set()
        try:
            os.killpg(process.pid, signal.SIGKILL)
        except Exception as e:
            logger.error("Exception trying to kill process: %s" % e)


def start_timer(p, timeout):
    """ Function to construct and return the timer thread and timed_out
    Args:
        p (object): This is the subprocess object which will be used to run the command of interest
        timeout (int): This is the timeout in seconds after which the command will be killed
    """
    # Start the background thread to catch timeout events
    timed_out = threading.Event()
    timer = threading.Timer(timeout, __timeout_func, args=(p, timed_out))
    timer.daemon = True
    if timeout is not None:
        timer.start()
    return timer, timed_out


def update_thread(pipes, thread_output, output_key, require_output):
    """ Function to construct and return background thread used to read a pickled object into the thread_output for updating
        the environment after executing a users code
        Args:
            started_threads (list): List containing background threads which have been started
            pipes (tuple): Tuple containing (read_pipe, write_pipe) which is the pipe the pickled obj is written to
            thread_output (dict): Dictionary containing the thread outputs which are used after executing the command
            output_key (str): Used to know where in the thread_output to store the output of this thread
            require_output (bool): Does the reader require valid pickled output.
    """
    ev = threading.Thread(target=__reader, args=(pipes, thread_output, output_key, require_output))
    ev.daemon = True
    ev.start()
    return ev


def execute(command,
            timeout=None,
            env=None,
            cwd=None,
            shell=True,
            python_setup='',
            eval_includes=None,
            update_env=False,
            ):
    """
    Execute an external command.
    This will execute an external python command when shell=False or an external bash command when shell=True
    Args:
        command (str): This is the command that we want to execute in string format
        timeout (int): This is the timeout which we want to assign to a function and it will be killed if it runs for longer than n seconds
        env (dict): This is the environment to use for launching the new command
        cwd (str): This is the cwd the command is to be executed within.
        shell (bool): True for a bash command to be executed, False for a command to be executed within Python
        python_setup (str): A python command to be executed beore the main command is
        eval_includes (str): An string used to construct an environment which, if passed, is used to eval the stdout into a python object
        update_env (bool): Should we update the env being passed to what the env was after the command finished running
    """

    if update_env and env is None:
        raise GangaException('Cannot update the environment if None given.')

    if not shell:
        # We want to run a python command inside a small Python wrapper
        stream_command = 'python -'
        command, pkl_file_pipes, env_file_pipes = python_wrapper(command, python_setup, update_env)
    else:
        # We want to run a shell command inside a _NEW_ shell environment.
        # i.e. What I run here I expect to behave in the same way from the command line after I exit Ganga
        stream_command = "bash "
        if update_env:
            # note the exec gets around the problem of indent and base64 gets
            # around the \n
            command_update, env_file_pipes = env_update_script()
            command += ''';python -c "import base64;exec(base64.b64decode(%s))"''' % base64.b64encode(command_update.encode("utf-8"))

    # Some minor changes to cleanup the getting of the env
    if env is None:
        env = os.environ

    # Construct the object which will contain the environment we want to run the command in
    p = subprocess.Popen(stream_command, shell=True, env=env, cwd=cwd, preexec_fn=os.setsid,
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         universal_newlines=True, close_fds=False)

    # This is where we store the output
    thread_output = {}

    # Start the timer thread used to kill commands which have likely stalled
    timer, timed_out = start_timer(p, timeout)

    if update_env:
        env_output_key = 'env_output'
        update_env_thread = update_thread(env_file_pipes, thread_output, env_output_key, require_output=True)
    if not shell:
        pkl_output_key = 'pkl_output'
        update_pkl_thread = update_thread(pkl_file_pipes, thread_output, pkl_output_key, require_output=False)

    # Execute the main command of interest
    logger.debug("Executing Command:\n'%s'" % str(command))
    stdout, stderr = p.communicate(command)

    # Close the timeout watching thread
    logger.debug("stdout: %s" % stdout)
    logger.debug("stderr: %s" % stderr)

    timer.cancel()
    if timeout is not None:
        timer.join()

    # Finish up and decide what to return
    if stderr != '':
        # this is still debug as using the environment from dirac default_env maked a stderr message dump out
        # even though it works
        logger.debug(stderr)

    if timed_out.isSet():
        return 'Command timed out!'

    # Decode any pickled objects from disk
    if update_env:
        update_env_thread.join()
        if env_output_key in thread_output:
            env.update(thread_output[env_output_key])
        else:
            logger.error("Expected to find the updated env after running a command")
            logger.error("Command: %s" % command)
            logger.error("stdout: %s" % stdout)
            logger.error("stderr: %s" % stderr)
            raise RuntimeError("Missing update env after running command")

    if not shell and not eval_includes:
        update_pkl_thread.join()
        if pkl_output_key in thread_output:
            return thread_output[pkl_output_key]

    stdout_temp = None
    try:
        # If output
        if stdout:
            if isinstance(stdout, bytes):
                stdout_temp = pickle.loads(stdout)
            else:
                try:
                    stdout_temp = pickle.loads(stdout.encode("utf-8"))
                except pickle.UnpicklingError:
                    stdout_temp = pickle.loads(stdout.encode("latin1"))
    # Downsides to wanting to be explicit in how this failed is you need to know all the ways it can!
    except (pickle.UnpicklingError, EOFError, ValueError) as err:
        if not shell:
            log = logger.error
        else:
            log = logger.debug
        log("Command Failed to Execute:\n%s" % command)
        log("Command Output is:\n%s" % stdout)
        log("Error received:\n%s" % err)

    if not stdout_temp:
        local_ns = locals()
        if isinstance(eval_includes, str):
            try:
                exec(eval_includes, {}, local_ns)
            except:
                logger.debug("Failed to eval the env, can't eval stdout")
                pass
        if isinstance(stdout, str) and stdout:
            try:
                stdout_temp = eval(stdout, {}, local_ns)
            except Exception as err2:
                logger.debug("Err2: %s" % str(err2))
                pass

    if stdout_temp:
        stdout = stdout_temp

    return stdout




from __future__ import absolute_import

from GangaCore.Core.exceptions import PluginError
raise PluginError("The GangaLSST module has not been upgraded for python 3. The last python 2 Ganga version is 7.1.15 . Please contact the ganga devs to discuss updating this module.")

from . import Lib




import datetime

from GangaCore.Core.exceptions import GangaTypeError
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Base.Proxy import stripProxy
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem

import GangaCore.Utility.Config

from GangaCore.Utility.logging import getLogger
GangaCore.Utility.Config.config_scope['datetime'] = datetime
logger = getLogger(modulename=True)


class JobTime(GangaObject):

    """Job timestamp access.
       In development

       Changes in the status of a Job are timestamped - a datetime object
       is stored in the dictionary named 'timestamps', in Coordinated
       Universal Time(UTC). More information on datetime objects can be
       found at:

       http://docs.python.org/library/datetime.html

       Datetime objects can be subtracted to produce a 'timedelta' object.
       More information about these can be found at the above address.
       '+', '*', and '/' are not supported by datetime objects.

       Datetime objects can be formatted into strings using the
       .strftime(format_string) application, and the strftime codes.
       e.g. %Y -> year as integer
            %a -> abbreviated weekday name
            %M -> minutes as inetger

       The full list can be found at:
       http://docs.python.org/library/datetime.html#strftime-behavior

       Standard status types with built in access methods are:
       -'new'
       -'submitted'
       -'running'
       -'completed'
       -'killed'
       -'failed'

       These return a string with default format %Y/%m/%d @ %H:%M:%S. A
       custom format can be specified in the arguement.

       Any information stored within the timestamps dictionary can also be
       extracted in the way as in would be for a standard, non-application
       specific python dictionary.

       For a table display of the Job's timestamps use .time.display(). For
       timestamps details from the backend use .time.details()


    """

    timestamps = {}
    sj_statlist = []

    _schema = Schema(Version(0, 0), {'timestamps': SimpleItem(defvalue={}, doc="Dictionary containing timestamps for job", summary_print='_timestamps_summary_print')
                                     })

    _category = 'jobtime'
    _name = 'JobTime'
    _exportmethods = ['display',
                      'new',
                      'submitting',
                      'submitted',
                      'backend_running',
                      'backend_final',
                      'backend_completing',
                      'completing',
                      'final',
                      'running',
                      'runtime',
                      'waittime',
                      'submissiontime',
                      'details',
                      'printdetails']

    def __init__(self):
        super(JobTime, self).__init__()
        self.timestamps = {}
        # this makes sure the contents of the list don't get copied when the
        # Job does.
        self.sj_statlist = []

    def __deepcopy__(self, memo):
        obj = super(JobTime, self).__deepcopy__(memo)
        # Lets not re-initialize the object as we lose history from previous submissions
        # obj.newjob()
        return obj

    def newjob(self):
        """Timestamps job upon creation.
        """
        t = datetime.datetime.utcnow()
        self.timestamps['new'] = t
        # this makes sure the contents of the list don't get copied when the
        # Job does.
        self.sj_statlist = []

    def timenow(self, status):
        """Updates timestamps as job status changes.
        """
        j = self.getJobObject()
        t_now = datetime.datetime.utcnow()
        b_list = ['running', 'completing', 'completed', 'failed']
        final = ['killed', 'failed', 'completed']
        backend_final = ['failed', 'completed']
        ganga_master = ['new', 'submitting', 'killed']

        id = j.id
        if id is None:
            id = str("unknown")
        logger.debug("Job %s called timenow('%s')", str(id), status)

        # standard method:
        if not j.subjobs:
            # backend stamps
            if status in b_list:
                for childstatus in b_list:
                    be_statetime = stripProxy(j.backend).getStateTime(childstatus)
                    if be_statetime is not None:
                        if childstatus in backend_final:
                            self.timestamps["backend_final"] = be_statetime
                            logger.debug(
                                "Wrote 'backend_final' to timestamps.")
                        else:
                            self.timestamps[
                                "backend_" + childstatus] = be_statetime
                            logger.debug(
                                "Wrote 'backend_%s' to timestamps.", childstatus)
                    if childstatus == status:
                        break
            # ganga stamps
            if status in final:
                self.timestamps["final"] = t_now
                logger.debug("Wrote 'final' to timestamps.")
            else:
                self.timestamps[status] = t_now
                logger.debug("Wrote '%s' to timestamps.", status)

        # subjobs method:
        if j.master is not None:  # identifies subjobs
            logger.debug(
                "j.time.timenow() caught subjob %d.%d in the '%s' status", j.master.id, j.id, status)

            for written_status in list(j.time.timestamps.keys()):
                if written_status not in j.master.time.sj_statlist:
                    j.master.time.sj_statlist.append(written_status)
                    logger.debug(
                        "written_status: '%s' written to sj_statlist", written_status)

        # master job method
        if j.subjobs:  # identifies master job
            logger.debug(
                "j.time.timenow() caught master job %d in the '%s' status", j.id, status)
            if status in ganga_master:  # don't use subjob stamp for these
                self.timestamps[status] = t_now
                logger.debug(
                    "status: '%s' in ganga_master written to master timestamps.", status)
            else:
                for state in self.sj_statlist:
                    if state not in ganga_master:
                        j.time.timestamps[
                            state] = self.sjStatList_return(state)
                        logger.debug(
                            "state: '%s' from sj_statlist to written to master timestamps.", state)
                    else:
                        pass

    def sjStatList_return(self, status):
        list = []
        final = ['backend_final', 'final']
        j = self.getJobObject()
        for sjs in j.subjobs:
            try:
                if isinstance(sjs.time.timestamps[status], datetime.datetime):
                    list.append(sjs.time.timestamps[status])
                else:
                    logger.debug(
                        'Attempt to add a non datetime object in the timestamp, job=%d, subjob=%d', j.id, sjs.id)
            except KeyError:
                logger.debug(
                    "Status '%s' not found in timestamps of job %d.%d.", status, sjs.master.id, sjs.id)
        list.sort()
        try:
            if status in final:
                return list[-1]
            return list[0]
        except IndexError:
            # change this to a more appropriate debug.
            logger.debug(
                "IndexError: ID: %d, Status: '%s', length of list: %d", j.id, status, len(list))

    def display(self, format="%Y/%m/%d %H:%M:%S"):
        return self._display(format)

    # Justin 10.9.09: I think 'ljust' might be just as good if not better than
    # 'rjust' here:
    def _display(self, format="%Y/%m/%d %H:%M:%S", interactive=False):
        """Displays existing timestamps in a table.

           Format can be specified by typing a string of the appropriate strftime() behaviour codes as the arguement.
           e.g. '%H:%M:%S' ==> 13:55:01

           For a full list of codes see
           http://docs.python.org/library/datetime.html?#strftime-behavior
        """
        retstr = ''
        T = datetime.datetime.now()
        tstring = T.strftime(format)
        length = len(tstring)
        times = [0 for _ in self.timestamps.keys()]
        for i, (ts_key, ts_value) in enumerate(self.timestamps.items()):
            try:
                times[i] = ts_value.strftime(format).rjust(length) + ' - ' + ts_key
            except AttributeError:
                times[i] = str(ts_value).rjust(length) + ' - ' + ts_key

        # try to make chronological - can fail when timestamps are the same to
        # nearest sec -> becomes alphabetical...
        times.sort()
        retstr = retstr + '\n' + \
            'Time (UTC)'.rjust(length) + '   Status' + '\n'
        for i in range(0, 21):
            retstr = retstr + '- '
        retstr = retstr + '\n'
        for i in range(0, len(times)):
            retstr = retstr + times[i] + '\n'
        return retstr

    def _timestamps_summary_print(self, value, verbosity_level, interactive=False):
        """Used to display timestamps when JobTime object is displayed.
        """
        return self._display(interactive=interactive)

    # This didn't work:
    #
    # def __str__(self):
    #    """ string cast """
    #    return self._display()

    def details(self, subjob=None):
        """Obtains all timestamps available from the job's specific backend.

           Subjob arguement: None  = default
                             'all' = gets details for ALL SUBJOBS. You have been warned.
                             int   = gets details for subjob number 'int'

           No argument is required for a job with no subjobs.    
        """
        j = self.getJobObject()
        idstr = ''
        detdict = {}

        # If job is SUBJOB do the normal procedure. Not sure this clause is
        # neccessary as subjobs will be caught normally
        if j.master is not None:
            logger.debug("j.time.details(): subjob %d.%d caught.", j.master.id, j.id)
            detdict = j.backend.timedetails()
            return detdict

        # If job is MASTER iterate over subjobs and do normal method. This
        # isn't going to be ideal for a large number of subjobs
        if j.subjobs:
            logger.debug("j.time.details(): master job %d caught.", j.id)
            idstr = str(j.id)

            # User wants 'all'
            if subjob == 'all':
                keyin = None

                # NOTE: The interactive loop below was more an exercise for learning how 'keyin' is used than a useful addition.
                # ask whether user really wants to print timedetails for all
                # their jobs:
                while keyin is None:
                    keyin = input("Are you sure you want details for ALL %d subjobs(y/n)?" % len(j.subjobs))
                    # if yes carry on at for loop
                    if keyin.lower() == 'y':
                        pass
                    # if no return None. Doesn't execute rest of method
                    elif keyin.lower() == 'n':
                        return None
                    # if something else - asks again
                    else:
                        logger.info("y/n please!")
                        keyin = None

                for jobs in j.subjobs:
                    subidstr = idstr + '.' + str(jobs.id)
                    # String needs more info if it is going to stay in.
                    logger.debug(
                        "Subjob: %d, Backend ID: %d", jobs.id, jobs.backend.id)
                    detdict[subidstr] = jobs.backend.timedetails()
                return detdict

            # no arguement specified
            elif subjob is None:
                logger.debug(
                    "j.time.details(): no subjobs specified for this master job.")
                return None

            # Subjob id or string passed
            else:
                # string = error
                if not isinstance(subjob, int):
                    raise GangaTypeError("Subjob id requires type 'int'")
                # subjob id supplied
                for sj in j.subjobs:
                    if sj.id == subjob:
                        logger.debug(
                            "Subjob: %d, Backend ID: %d", sj.id, sj.backend.id)
                        detdict = sj.backend.timedetails()
                        return detdict
                    else:
                        pass
                if subjob >= len(j.subjobs):
                    logger.warning(
                        "Index '%s' is out of range. Corresponding subjob does not exist.", str(subjob))
                    return None

            logger.debug(
                "subjob arguement '%s' has failed to be caught and dealt with.", subjob)
            return None

        detdict = j.backend.timedetails()  # called if no subjobs
        return detdict

    def printdetails(self, subjob=None):
        """Prints backend details to screen by calling details() and printing the returned dictionary.
        """
        j = self.getJobObject()
        if subjob == 'all':
            # the warning and action taken below are pretty annoying, but I was
            # unsure how to deal with the request to print the details for all
            # n subjobs, which seems unlikely to be made.
            logger.warning(
                "It might be unwise to print all subjobs details. Use details() and extract relevant info from dictionary.")
            return None
        pd = self.details(subjob)
        for key in pd.keys():
            logger.info(key, '\t', pd[key])

    def runtime(self):
        """Method which returns the 'runtime' of the specified job.

           The runtime is calculated as the duration between the job entering the 'running' state and the job entering the 'completed' state.
        """
        end_list = ['killed', 'completed', 'failed']
        end_stamps = {}

        # if master job, sum:
        j = self.getJobObject()
        if j.subjobs:
            masterrun = datetime.timedelta(0, 0, 0)
            for jobs in j.subjobs:
                masterrun = masterrun + jobs.time.runtime()
            return masterrun
        # all other jobs:
        return self.duration('backend_running', 'backend_final')

    def waittime(self):
        """Method which returns the waiting time of the specified job.

           The waiting time is calculated as the duration between the job entering the 'submitted' state and entering the 'running' state.
        """
        # master job:
        j = self.getJobObject()
        if j.subjobs:
            start_list = []
            end_list = []
            for jobs in j.subjobs:
                start_list.append(jobs.time.timestamps['submitted'])
                end_list.append(jobs.time.timestamps['backend_running'])
            start_list.sort()
            end_list.sort()
            start = start_list[0]
            end = end_list[len(end_list) - 1]
            masterwait = end - start
            return masterwait
        # all other jobs:
        return self.duration('submitted', 'backend_running')

    def submissiontime(self):
        """Method which returns submission time of specified job.

           Calculation: sub_time =  submitted - submitting.
        """
        j = self.getJobObject()
        if j.subjobs:
            start_list = []
            end_list = []
            for jobs in j.subjobs:
                end_list.append(jobs.time.timestamps['submitted'])
            end_list.sort()
            start = j.time.timestamps['submitting']
            end = end_list[len(end_list) - 1]
            mastersub = end - start
            return mastersub
        return self.duration('submitting', 'submitted')

    def duration(self, start, end):
        """Returns duration between two specified timestamps as timedelta object.
        """
        if start in self.timestamps:
            if end in self.timestamps:
                s, e = self.timestamps[start], self.timestamps[end]
                s_micro, e_micro = datetime.timedelta(
                    0, 0, s.microsecond), datetime.timedelta(0, 0, e.microsecond)
                e, s = e - e_micro, s - s_micro
                td = e - s

                # method for rounding removed because timestamps aren't always recorded with microsecond precision, and stamping accuracy isn't high enough to justify doing so
#                ds = td.days
 #               secs = td.seconds
#                micros = td.microseconds
 #               if micros >= 500000:
  #                  secs +=1

                dur = td  # datetime.timedelta(days=ds, seconds=secs)
                return dur
            else:
                logger.warning(
                    "Could not calculate duration: '%s' not found.", end)
        else:
            logger.warning(
                "Could not calculate duration: '%s' not found.", start)
        return None

    def statetime(self, status, format=None):
        """General method for obtaining the specified timestamp in specified format.
        """
        if status not in self.timestamps:
            logger.debug("Timestamp '%s' not available.", status)
            return None
        if format is not None:
            return self.timestamps[status].strftime(format)
        return self.timestamps[status]

    def new(self, format=None):
        """Method for obtaining 'new' timestamp.
        """
        return self.statetime('new', format)

    def submitting(self, format=None):
        """Method for obtaining 'submitting' timestamp.
        """
        return self.statetime('submitting', format)

    def submitted(self, format=None):
        """Method for obtaining 'submitted' timestamp.
        """
        return self.statetime('submitted', format)

    def backend_running(self, format=None):
        """Method for obtaining 'backend_running' timestamp.
        """
        return self.statetime('backend_running', format)

    def backend_final(self, format=None):
        """Method for obtaining 'backend_final' timestamp.
        """
        return self.statetime('backend_final', format)

    def backend_completing(self, format=None):
        """Method for obtaining 'backend_completing' timestamp.
        """
        return self.statetime('backend_completing', format)

    def completing(self, format=None):
        """Method for obtaining 'completing' timestamp.
        """
        return self.statetime('completing', format)

    def final(self, format=None):
        """Method for obtaining 'final' timestamp.
        """
        return self.statetime('final', format)

    def running(self, format=None):
        """Method for obtaining 'running' timestamp.
        """
        return self.statetime('running', format)



"""Module for sharing code between the Runtime and Download handlers."""
import GangaCore.Utility.logging
import copy
logger = GangaCore.Utility.logging.getLogger()


def findPythonVersion(rootsys):
    '''Digs around in rootsys for config files and then greps
    the version of python used'''
    import os

    def lookInFile(config):
        '''Looks in the specified file for the build config
        and picks out the python version'''
        version = None
        if os.path.exists(config):
            # loop through the file looking for #define
            for line in open(config):
                if line.startswith('#define R__CONFIGUREOPTION'):
                    for arg in line.split(' '):  # look at value of #define
                        if arg.startswith('PYTHONDIR'):
                            arglist = arg.split('/')
                            # TODO: Would like some kind of check for arch here
                            if len(arglist) > 1:  # prevent index out of range
                                version = arglist[-2]
        return version

    def useRootConfig(rootsys):
        """Use the new root-config features to find the python version"""
        version = None
        root_config = os.path.join(rootsys, 'bin', 'root-config')
        if os.path.exists(root_config):
            import subprocess

            args = [root_config, '--python-version']

            run = subprocess.Popen(
                ' '.join(args), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = [e.splitlines() for e in run.communicate()]
            code = run.returncode
            if code == 0 and out and not err and len(out) == 1:
                split = out[0].split('.')
                if len(out) != len(split):
                    version = '.'.join(split)
        return version

    version = None
    for f in ['RConfigOptions.h', 'config.h', 'RConfigure.h']:
        version = lookInFile(os.path.join(rootsys, 'include', f))
        if version is not None:
            break
    if version is None:
        version = useRootConfig(rootsys)
    return version


def setEnvironment(key, value, update=False, environment=None):
    '''Sets an environment variable. If update=True, it prepends it to
    the current value with os.pathsep as the seperator.'''
    import os
    if environment is None:
        environment = copy.deepcopy(os.environ)

    if update and key in environment:
        value += (os.pathsep + environment[key])  # prepend
    environment[key] = value

    return environment



"""
    Makes the necesary files to convert from plain html of
    Python 1.5 and 1.5.x Documentation to
    Microsoft HTML Help format version 1.1
    Doesn't change the html's docs.

    by hernan.foffani@iname.com
    no copyright and no responsabilities.

    modified by Dale Nagata for Python 1.5.2

    Renamed from make_chm.py to prechm.py, and checked into the Python
    project, 19-Apr-2002 by Tim Peters.  Assorted modifications by Tim
    and Fred Drake.  Obtained from Robin Dunn's .chm packaging of the
    Python 2.2 docs, at <http://alldunn.com/ganga/>.
"""

import sys
import os
from formatter import NullWriter, AbstractFormatter
from htmllib import HTMLParser
import getopt
import cgi

usage_mode = '''
Usage: make_chm.py [-c] [-k] [-p] [-v 1.5[.x]] filename
    -c: does not build filename.hhc (Table of Contents)
    -k: does not build filename.hhk (Index)
    -p: does not build filename.hhp (Project File)
    -v 1.5[.x]: makes help for the python 1.5[.x] docs
        (default is python 1.5.2 docs)
'''

# Project file (*.hhp) template.  'arch' is the file basename (like
# the pythlp in pythlp.hhp); 'version' is the doc version number (like
# the 2.2 in Python 2.2).
# The magical numbers in the long line under [WINDOWS] set most of the
# user-visible features (visible buttons, tabs, etc).
# About 0x10384e:  This defines the buttons in the help viewer.  The
# following defns are taken from htmlhelp.h.  Not all possibilities
# actually work, and not all those that work are available from the Help
# Workshop GUI.  In particular, the Zoom/Font button works and is not
# available from the GUI.  The ones we're using are marked with 'x':
#
#    0x000002   Hide/Show   x
#    0x000004   Back        x
#    0x000008   Forward     x
#    0x000010   Stop
#    0x000020   Refresh
#    0x000040   Home        x
#    0x000080   Forward
#    0x000100   Back
#    0x000200   Notes
#    0x000400   Contents
#    0x000800   Locate      x
#    0x001000   Options     x
#    0x002000   Print       x
#    0x004000   Index
#    0x008000   Search
#    0x010000   History
#    0x020000   Favorites
#    0x040000   Jump 1
#    0x080000   Jump 2
#    0x100000   Zoom/Font   x
#    0x200000   TOC Next
#    0x400000   TOC Prev

project_template = '''
[OPTIONS]
Compiled file=%(arch)s.chm
Contents file=%(arch)s.hhc
Default Window=%(arch)s
Default topic=index.html
Display compile progress=No
Full text search stop list file=%(arch)s.stp
Full-text search=Yes
Index file=%(arch)s.hhk
Language=0x409
Title=Python %(version)s Documentation

[WINDOWS]
%(arch)s="Python %(version)s Documentation","%(arch)s.hhc","%(arch)s.hhk",\
"index.html","index.html",,,,,0x63520,220,0x10384e,[0,0,1024,768],,,,,,,0

[FILES]
'''

contents_header = '''\
<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML//EN">
<HTML>
<HEAD>
<meta name="GENERATOR" content="Microsoft&reg; HTML Help Workshop 4.1">
<!-- Sitemap 1.0 -->
</HEAD><BODY>
<OBJECT type="text/site properties">
	<param name="Window Styles" value="0x801227">
	<param name="ImageType" value="Folder">
</OBJECT>
<UL>
'''

contents_footer = '''\
</UL></BODY></HTML>
'''

object_sitemap = '''\
<OBJECT type="text/sitemap">
    <param name="Name" value="%s">
    <param name="Local" value="%s">
</OBJECT>
'''

# List of words the full text search facility shouldn't index.  This
# becomes file ARCH.stp.  Note that this list must be pretty small!
# Different versions of the MS docs claim the file has a maximum size of
# 256 or 512 bytes (including \r\n at the end of each line).
# Note that "and", "or", "not" and "near" are operators in the search
# language, so no point indexing them even if we wanted to.
stop_list = '''
a  and  are  as  at
be  but  by
for
if  in  into  is  it
near  no  not
of  on  or
such
that  the  their  then  there  these  they  this  to
was  will  with
'''

# s is a string or None.  If None or empty, return None.  Else tack '.html'
# on to the end, unless it's already there.
def addhtml(s):
    if s:
        if not s.endswith('.html'):
            s += '.html'
    return s

# Convenience class to hold info about "a book" in HTMLHelp terms == a doc
# directory in Python terms.
class Book:
    def __init__(self, directory, title, firstpage,
                 contentpage=None, indexpage=None):
        self.directory   = directory
        self.title       = title
        self.firstpage   = addhtml(firstpage)
        self.contentpage = addhtml(contentpage)
        self.indexpage   = addhtml(indexpage)

# Library Doc list of books:
# each 'book' : (Dir, Title, First page, Content page, Index page)
supported_libraries = {
    '2.3':
    [
        Book('.', 'Main page', 'index'),
        Book('.', 'Global Module Index', 'modindex'),
        Book('whatsnew', "What's New", 'index', 'contents'),
        Book('tut','Tutorial','tut','node2'),
        Book('lib','Library Reference','lib','contents','genindex'),
        Book('ref','Language Reference','ref','contents','genindex'),
        Book('mac','Macintosh Reference','mac','contents','genindex'),
        Book('ext','Extending and Embedding','ext','contents'),
        Book('api','Python/C API','api','contents','genindex'),
        Book('doc','Documenting Python','doc','contents'),
        Book('inst','Installing Python Modules', 'inst', 'index'),
        Book('dist','Distributing Python Modules', 'dist', 'index'),
    ],

    '2.2':
    [
        Book('.', 'Main page', 'index'),
        Book('.', 'Global Module Index', 'modindex'),
        Book('whatsnew', "What's New", 'index', 'contents'),
        Book('tut','Tutorial','tut','node2'),
        Book('lib','Library Reference','lib','contents','genindex'),
        Book('ref','Language Reference','ref','contents','genindex'),
        Book('mac','Macintosh Reference','mac','contents','genindex'),
        Book('ext','Extending and Embedding','ext','contents'),
        Book('api','Python/C API','api','contents','genindex'),
        Book('doc','Documenting Python','doc','contents'),
        Book('inst','Installing Python Modules', 'inst', 'index'),
        Book('dist','Distributing Python Modules', 'dist', 'index'),
    ],

    '2.1.1':
    [
        Book('.', 'Main page', 'index'),
        Book('.', 'Global Module Index', 'modindex'),
        Book('tut','Tutorial','tut','node2'),
        Book('lib','Library Reference','lib','contents','genindex'),
        Book('ref','Language Reference','ref','contents','genindex'),
        Book('mac','Macintosh Reference','mac','contents','genindex'),
        Book('ext','Extending and Embedding','ext','contents'),
        Book('api','Python/C API','api','contents','genindex'),
        Book('doc','Documenting Python','doc','contents'),
        Book('inst','Installing Python Modules', 'inst', 'index'),
        Book('dist','Distributing Python Modules', 'dist', 'index'),
    ],

    '2.0.0':
    [
        Book('.', 'Global Module Index', 'modindex'),
        Book('tut','Tutorial','tut','node2'),
        Book('lib','Library Reference','lib','contents','genindex'),
        Book('ref','Language Reference','ref','contents','genindex'),
        Book('mac','Macintosh Reference','mac','contents','genindex'),
        Book('ext','Extending and Embedding','ext','contents'),
        Book('api','Python/C API','api','contents','genindex'),
        Book('doc','Documenting Python','doc','contents'),
        Book('inst','Installing Python Modules', 'inst', 'contents'),
        Book('dist','Distributing Python Modules', 'dist', 'contents'),
    ],

    # <dnagata@creo.com> Apr 17/99: library for 1.5.2 version:
    # <hernan.foffani@iname.com> May 01/99: library for 1.5.2 (04/30/99):
    '1.5.2':
    [
        Book('tut','Tutorial','tut','node2'),
        Book('lib','Library Reference','lib','contents','genindex'),
        Book('ref','Language Reference','ref','contents','genindex'),
        Book('mac','Macintosh Reference','mac','contents','genindex'),
        Book('ext','Extending and Embedding','ext','contents'),
        Book('api','Python/C API','api','contents','genindex'),
        Book('doc','Documenting Python','doc','contents')
    ],

    # library for 1.5.1 version:
    '1.5.1':
    [
        Book('tut','Tutorial','tut','contents'),
        Book('lib','Library Reference','lib','contents','genindex'),
        Book('ref','Language Reference','ref-1','ref-2','ref-11'),
        Book('ext','Extending and Embedding','ext','contents'),
        Book('api','Python/C API','api','contents','genindex')
    ],

    # library for 1.5 version:
    '1.5':
    [
        Book('tut','Tutorial','tut','node1'),
        Book('lib','Library Reference','lib','node1','node268'),
        Book('ref','Language Reference','ref-1','ref-2','ref-11'),
        Book('ext','Extending and Embedding','ext','node1'),
        Book('api','Python/C API','api','node1','node48')
    ]
}

# AlmostNullWriter doesn't print anything; it just arranges to save the
# text sent to send_flowing_data().  This is used to capture the text
# between an anchor begin/end pair, e.g. for TOC entries.

class AlmostNullWriter(NullWriter):

    def __init__(self):
        NullWriter.__init__(self)
        self.saved_clear()

    def send_flowing_data(self, data):
        stripped = data.strip()
        if stripped:    # don't bother to save runs of whitespace
            self.saved.append(stripped)

    # Forget all saved text.
    def saved_clear(self):
        self.saved = []

    # Return all saved text as a string.
    def saved_get(self):
        return ' '.join(self.saved)

class HelpHtmlParser(HTMLParser):

    def __init__(self, formatter, path, output):
        HTMLParser.__init__(self, formatter)
        self.path = path    # relative path
        self.ft = output    # output file
        self.indent = 0     # number of tabs for pretty printing of files
        self.proc = False   # True when actively processing, else False
                            # (headers, footers, etc)
        # XXX This shouldn't need to be a stack -- anchors shouldn't nest.
        # XXX See SF bug <http://www.python.org/sf/546579>.
        self.hrefstack = [] # stack of hrefs from anchor begins

    def begin_group(self):
        self.indent += 1
        self.proc = True

    def finish_group(self):
        self.indent -= 1
        # stop processing when back to top level
        self.proc = self.indent > 0

    def anchor_bgn(self, href, name, type):
        if self.proc:
            # XXX See SF bug <http://www.python.org/sf/546579>.
            # XXX index.html for the 2.2.1 language reference manual contains
            # XXX nested <a></a> tags in the entry for the section on blank
            # XXX lines.  We want to ignore the nested part completely.
            if len(self.hrefstack) == 0:
                self.saved_clear()
                self.hrefstack.append(href)

    def anchor_end(self):
        if self.proc:
            # XXX See XXX above.
            if self.hrefstack:
                title = cgi.escape(self.saved_get(), True)
                path = self.path + '/' + self.hrefstack.pop()
                self.tab(object_sitemap % (title, path))

    def start_dl(self, atr_val):
        self.begin_group()

    def end_dl(self):
        self.finish_group()

    def do_dt(self, atr_val):
        # no trailing newline on purpose!
        self.tab("<LI>")

    # Write text to output file.
    def write(self, text):
        self.ft.write(text)

    # Write text to output file after indenting by self.indent tabs.
    def tab(self, text=''):
        self.write('\t' * self.indent)
        if text:
            self.write(text)

    # Forget all saved text.
    def saved_clear(self):
        self.formatter.writer.saved_clear()

    # Return all saved text as a string.
    def saved_get(self):
        return self.formatter.writer.saved_get()

class IdxHlpHtmlParser(HelpHtmlParser):
    # nothing special here, seems enough with parent class
    pass

class TocHlpHtmlParser(HelpHtmlParser):

    def start_dl(self, atr_val):
        self.begin_group()
        self.tab('<UL>\n')

    def end_dl(self):
        self.finish_group()
        self.tab('</UL>\n')

    def start_ul(self, atr_val):
        self.begin_group()
        self.tab('<UL>\n')

    def end_ul(self):
        self.finish_group()
        self.tab('</UL>\n')

    def do_li(self, atr_val):
        # no trailing newline on purpose!
        self.tab("<LI>")

def index(path, indexpage, output):
    parser = IdxHlpHtmlParser(AbstractFormatter(AlmostNullWriter()),
                              path, output)
    f = open(path + '/' + indexpage)
    parser.feed(f.read())
    parser.close()
    f.close()

def content(path, contentpage, output):
    parser = TocHlpHtmlParser(AbstractFormatter(AlmostNullWriter()),
                              path, output)
    f = open(path + '/' + contentpage)
    parser.feed(f.read())
    parser.close()
    f.close()

def do_index(library, output):
    output.write('<UL>\n')
    for book in library:
        print '\t', book.title, '-', book.indexpage
        if book.indexpage:
            index(book.directory, book.indexpage, output)
    output.write('</UL>\n')

def do_content(library, version, output):
    output.write(contents_header)
    for book in library:
        print '\t', book.title, '-', book.firstpage
        path = book.directory + "/" + book.firstpage
        output.write('<LI>')
        output.write(object_sitemap % (book.title, path))
        if book.contentpage:
            content(book.directory, book.contentpage, output)
    output.write(contents_footer)

# Fill in the [FILES] section of the project (.hhp) file.
# 'library' is the list of directory description tuples from
# supported_libraries for the version of the docs getting generated.
def do_project(library, output, arch, version):
    output.write(project_template % locals())
    pathseen = {}
    for book in library:
        directory = book.directory
        path = directory + '\\%s\n'
        for page in os.listdir(directory):
            if page.endswith('.html') or page.endswith('.css'):
                fullpath = path % page
                if fullpath not in pathseen:
                    output.write(fullpath)
                    pathseen[fullpath] = True

def openfile(file):
    try:
        p = open(file, "w")
    except IOError, msg:
        print file, ":", msg
        sys.exit(1)
    return p

def usage():
        print usage_mode
        sys.exit(0)

def do_it(args = None):
    if not args:
        args = sys.argv[1:]

    if not args:
        usage()

    try:
        optlist, args = getopt.getopt(args, 'ckpv:')
    except getopt.error, msg:
        print msg
        usage()

    if not args or len(args) > 1:
        usage()
    arch = args[0]

    version = None
    for opt in optlist:
        if opt[0] == '-v':
            version = opt[1]
            break
    if not version:
        usage()

    library = supported_libraries[version]

    if not (('-p','') in optlist):
        fname = arch + '.stp'
        f = openfile(fname)
        print "Building stoplist", fname, "..."
        words = stop_list.split()
        words.sort()
        for word in words:
            print >> f, word
        f.close()

        f = openfile(arch + '.hhp')
        print "Building Project..."
        do_project(library, f, arch, version)
        if version == '2.0.0':
            for image in os.listdir('icons'):
                f.write('icons'+ '\\' + image + '\n')

        f.close()

    if not (('-c','') in optlist):
        f = openfile(arch + '.hhc')
        print "Building Table of Content..."
        do_content(library, version, f)
        f.close()

    if not (('-k','') in optlist):
        f = openfile(arch + '.hhk')
        print "Building Index..."
        do_index(library, f)
        f.close()

if __name__ == '__main__':
    do_it()






import logging

from GangaCore.Utility.logging import getLogger, _formats


def test_logging():
    logger = getLogger()

    logger.info('info test')
    logger.error('error test')
    logger.warning('warning test')

    private_logger = logging.getLogger("TESTLOGGER.CHILD.GRANDCHILD")
    formatter = logging.Formatter(_formats['DEBUG'])
    console = logging.StreamHandler()
    console.setFormatter(formatter)
    private_logger.setLevel(logging.DEBUG)
    private_logger.addHandler(console)
    private_logger.critical('hello')



"""
Provides Persistency and the base class for all Registries
Also, a list of all Registries is kept here
"""

from GangaCore.Core.GangaRepository.GangaRepository import GangaRepository
from GangaCore.Core.exceptions import RepositoryError, InaccessibleObjectError, SchemaVersionError
from GangaCore.Core.GangaRepository.Registry import RegistryError, RegistryAccessError, RegistryKeyError, RegistryLockError, ObjectNotInRegistryError
from GangaCore.Core.GangaRepository import GangaRepositoryXML

allRegistries = {}


def addRegistry(registry):
    """ Add a registry to the global dict"""
    allRegistries[registry.name] = registry

def getRegistries():
    """ Get all registries from the global dict"""
    return list(allRegistries.values())

def getRegistry(name):
    """ Get a specific Registry from the global dict"""
    return allRegistries[name]

def getRegistrySlice(name):
    """ Get The registry slice wrapping the registry of choice"""
    return allRegistries[name].getSlice()

def getRegistryProxy(name):
    """ Get the proxied registry slice wrapping the registry of choice"""
    return allRegistries[name].getProxy()




###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PACKAGE.py,v 1.16 2009-06-18 08:32:28 dvanders Exp $
###############################################################################
""" Refer to Ganga/PACKAGE.py for details on the purpose of this module.
"""
#The platform needed by external packages. 
#Change this is you want to use external packages for different platform

import sys, os
from GangaCore.Utility.Setup import PackageSetup, checkPythonVersion, getExternalHome


_external_packages = { 
    'DQ2Clients' : { 'version' : '2.6.1_rc15',
                     'DQ2_HOME' : 'opt/dq2',
                     'PATH' : ['opt/dq2/bin','nordugrid/bin'],
                     'syspath' : ['opt/dq2/lib/','external/mysqldb32/'],
                     'LD_LIBRARY_PATH' : ['external/mysql32/','external/mysqldb32/','external/nordugrid/lib/'],
                     'DQ2_ENDUSER_SETUP' : 'True',
                     'noarch':True ,
                     'RUCIO_APPID' : 'ganga',
                     },
    'rucio-clients' : { 'version' : '1.11.3',
                     'PATH' : ['bin/'],
                     'syspath' : [ 'externals/kerberos/lib.slc6-x86_64-2.6', 'externals/kerberos/lib.slc6-i686-2.6', 'lib/python2.7/site-packages' ],
                     'RUCIO_HOME' : '/afs/cern.ch/sw/ganga/external/rucio-clients/0.2.13/noarch/', # done properly below
                     'RUCIO_AUTH_TYPE' : 'x509_proxy',
                     'RUCIO_ACCOUNT' : 'ganga',
                     'noarch':True ,
                     'RUCIO_APPID' : 'ganga',
                     },

    'panda-client' : { 'version' : '0.5.94',
                       'syspath':['lib/python2.6/site-packages'],
                       'CONFIGEXTRACTOR_PATH':'etc/panda/share',
                       'PANDA_SYS':'.',
                       'noarch':True
                       },
    'zsi' : { 'version' : '2.1-a1',  # Needed for pyAMI
                       'syspath':['lib/python'],
                       'noarch':True
                       },
    '4Suite' : { 'version' : '1.0.2.1', 
                 'PYTHONPATH':['lib/python2.4/site-packages'],
                 'syspath':['lib/python2.4/site-packages']
                       },
    'pyAMI' : { 'version' : '3.1.2.1', 
                       'syspath':['.'],
                       'noarch':True
                       }

    }

# use DQ2Clients 2.3.0 if running <2.6
import sys
if sys.hexversion < 0x2050000:
    _external_packages['DQ2Clients']['version'] = '2.3.0'

# use appropriate RUCIO Client version
_external_packages['rucio-clients']['RUCIO_HOME'] = os.path.join(getExternalHome(), 'rucio-clients', _external_packages['rucio-clients']['version'], 'noarch')
if 'CMTCONFIG' in os.environ and 'slc5' in os.environ['CMTCONFIG'] and 'i686' in os.environ['CMTCONFIG']:
    _external_packages['rucio-clients']['syspath'] = [ 'externals/kerberos/lib.slc6-i686-2.6', 'externals/kerberos/lib.slc6-x86_64-2.6', 'lib/python2.6/site-packages' ]
elif 'CMTCONFIG' in os.environ and 'slc5' in os.environ['CMTCONFIG']:
    _external_packages['rucio-clients']['syspath'] = [ 'externals/kerberos/lib.slc6-x86_64-2.6', 'externals/kerberos/lib.slc6-i686-2.6', 'lib/python2.6/site-packages' ]

# use python 2.7 versions if required
if sys.version_info[0] == 2 and sys.version_info[1] == 7:
    new_paths = []
    for p in _external_packages['rucio-clients']['syspath']:
        new_paths.append( p.replace('2.6', '2.7') )
    _external_packages['rucio-clients']['syspath'] = new_paths

# if Panda is already setup, use that
for p in sys.path:
    if p.find("PandaClient") > -1:
        del _external_packages['panda-client']
        break

setup = PackageSetup(_external_packages)

# Default minimum Python version number asked for by Ganga
_defaultMinVersion = "2.3"
_defaultMinHexVersion = 0x20300f0

def standardSetup(setup=setup):

    # here we assume that the Ganga has been already prepended to sys.path by the caller
    if checkPythonVersion(_defaultMinVersion,_defaultMinHexVersion):
        for name in setup.packages:
            if name == '4Suite' and (sys.hexversion > 0x2050000 and sys.hexversion < 0x2060000):
                # hack the 4Suite path for 2.5
                setup.packages['4Suite']['PYTHONPATH']  =  [ package.replace('2.4','2.5') for package in setup.packages['4Suite']['PYTHONPATH'] ]
                setup.packages['4Suite']['syspath']  =  [ package.replace('2.4','2.5') for package in setup.packages['4Suite']['syspath'] ]
            elif name == '4Suite' and sys.hexversion > 0x2060000:
                # hack the 4Suite path for 2.6
                setup.packages['4Suite']['PYTHONPATH']  =  [ package.replace('2.4','2.6') for package in setup.packages['4Suite']['PYTHONPATH'] ]
                setup.packages['4Suite']['syspath']  =  [ package.replace('2.4','2.6') for package in setup.packages['4Suite']['syspath'] ]
    else:
        sys.exit()

    for p in setup.packages:
        setup.prependPath(p,'PYTHONPATH')
        setup.prependPath(p,'LD_LIBRARY_PATH')
        setup.prependPath(p,'PATH')
        setup.setPath(p,'DQ2_HOME')
        if 'DQ2_ENDUSER_SETUP' in setup.packages[p]:
            os.environ['DQ2_ENDUSER_SETUP'] = setup.packages[p]['DQ2_ENDUSER_SETUP']
        setup.setPath(p,'PANDA_SYS')
        setup.setPath(p,'RUCIO_HOME')
        if 'RUCIO_AUTH_TYPE' in setup.packages[p]:
            os.environ['RUCIO_AUTH_TYPE'] = setup.packages[p]['RUCIO_AUTH_TYPE']

    # update the syspath and PYTHONPATH for the packages
    for name in setup.packages:
        setup.setSysPath(name)




from GangaGUI.gui import app
from flask import request


@app.route("/")
def dashboard():
    return "Hello GangaGUI"


@app.route('/shutdown', methods=['POST'])
def shutdown():
    func = request.environ.get('werkzeug.server.shutdown')
    func()
    return 'Shutting down GUI...'
